/****************************************************************************/

#include "nfi.h"

// This local define is used to create a buffer used in the add file dialog
// box to allow multiple file selections. A buffer size of 16k (the default)
// will allow about 205 files to be selected at once, assuming they all have
// exactly 80 characters, or over 1300 if they are 12 characters.
#define ADD_FILENAME_BUFFER_SIZE 16384

/****************************************************************************/

// This is the message dispatching class for the add dialog box
class CAddDialogDispatcher : public CDispatcher {
public:
    CAddDialogDispatcher (handlerType Handler) : CDispatcher (Handler) {}

    LRESULT MsgInit (HWND, UINT, WPARAM, LPARAM);
    LRESULT MsgHelp (HWND, UINT, WPARAM, LPARAM);

    LRESULT CmdDone    (HWND, WORD, WORD, HWND);
    LRESULT CmdAddDir  (HWND, WORD, WORD, HWND);
    LRESULT CmdAddFile (HWND, WORD, WORD, HWND);
    LRESULT CmdRemove  (HWND, WORD, WORD, HWND);

MESSAGE_LIST;
COMMAND_LIST;
};

START_MESSAGE_LIST (CAddDialogDispatcher)
    MESSAGE (WM_HELP       , MsgHelp)
    MESSAGE (WM_INITDIALOG , MsgInit)
END_MESSAGE_LIST

START_COMMAND_LIST (CAddDialogDispatcher)
    COMMAND (IDOK        , CmdDone)
    COMMAND (IDCANCEL    , CmdDone)
    COMMAND (IDC_ADDDIR  , CmdAddDir)
    COMMAND (IDC_ADDFILE , CmdAddFile)
    COMMAND (IDC_REMOVE  , CmdRemove)
END_COMMAND_LIST

/****************************************************************************/

static CAddDialogDispatcher DialogDispatch (handlerNone);
static char SelectedDirToAdd[MAX_PATH]="";

/*****************************************************************************
 * Name         - AddProc
 * Purpose      - This handles messages for the add dialog box by using it's
 *                dispatcher class.
 * Arguments    - Standard window procedure dialog boxes
 * Return Value - Depends on the message
 ****************************************************************************/
LRESULT CALLBACK AddProc (HWND hDlg , UINT uMessage ,
                          WPARAM wParam, LPARAM lParam)
{
 return DialogDispatch.dispatchMessage (hDlg , uMessage , wParam , lParam);
}

/*****************************************************************************
 * Name         - CAddDialogDispatcher::MsgInit
 * Purpose      - This handles the WM_INITDIALOG message.
 * Arguments    - hDlg: Handle of the dialog box
 * Return Value - TRUE
 ****************************************************************************/
LRESULT CAddDialogDispatcher::MsgInit (HWND hDlg , UINT /*uMessage*/ ,
                                       WPARAM /*wParam*/ , LPARAM /*lParam*/)
{
 // Tell the button to use the bitmap we loaded.
 SendMessage (GetDlgItem (hDlg , IDOK) , BM_SETIMAGE , (WPARAM) IMAGE_BITMAP , (LPARAM) hOKBitmap);
 SendMessage (GetDlgItem (hDlg , IDCANCEL) , BM_SETIMAGE , (WPARAM) IMAGE_BITMAP , (LPARAM) hCANCELBitmap);
 SendMessage (GetDlgItem (hDlg , IDC_ADDFILE) , BM_SETIMAGE , (WPARAM) IMAGE_BITMAP , (LPARAM) hADDFILEBitmap);
 SendMessage (GetDlgItem (hDlg , IDC_ADDDIR) , BM_SETIMAGE , (WPARAM) IMAGE_BITMAP , (LPARAM) hADDDIRBitmap);
 SendMessage (GetDlgItem (hDlg , IDC_REMOVE) , BM_SETIMAGE , (WPARAM) IMAGE_BITMAP , (LPARAM) hREMOVEBitmap);

 // Check or uncheck the checkbox used to tell the program if it should
 // resurce directories while adding based on the setup informatiuon we have
 // for it
 CheckDlgButton (hDlg , IDC_RECURSEDIRS , Setup.RecurseDirs ? BST_CHECKED : BST_UNCHECKED);

 // Add all the path info to the dialog control.
 HWND hCtrl = GetDlgItem (hDlg , IDC_ADDDIRCOMBO);
 WINASSERT (hCtrl);

 // This recurses, adding all of the paths to the listbox portion of the
 // combobox.
 AddPathToComboBox (NFIPathInfo , hCtrl);

 // Select the directory that is supposed to be used as the default add
 // location. We do a find to see what index it is, then select that. If
 // we don't find it, just select the first thing.
 int Index = SendMessage (hCtrl , CB_FINDSTRINGEXACT ,
                          (WPARAM) -1 ,
                          (LPARAM) Setup.AddDir);
 if (Index == CB_ERR)
     Index = 0;
 SendMessage (hCtrl , CB_SETCURSEL , Index , 0);

 // Tell the text control of the combo box how big it should let a person
 // type
 SendMessage (hCtrl , CB_LIMITTEXT , MAX_PATH - 1 , 0);

 // Center the dialog over the application window
 CenterWindow (hDlg , GetWindow (hDlg , GW_OWNER));

 return TRUE;
}

/*****************************************************************************
 * Name         - CAddDialogDispatcher::MsgHelp
 * Purpose      - This handles the WM_HELP message.
 * Arguments    - hDlg: Handle of the dialog box
 * Return Value - TRUE
 ****************************************************************************/
LRESULT CAddDialogDispatcher::MsgHelp (HWND hDlg , UINT /*uMessage*/ ,
                                       WPARAM /*wParam*/ , LPARAM lParam)
{
 static DWORD HelpID[] = {
     IDC_ADDLISTBOX    , IDC_ADDLISTBOX    ,
     IDC_ADDFILE       , IDC_ADDFILE       ,
     IDC_ADDDIR        , IDC_ADDDIR        ,
     IDC_REMOVE        , IDC_REMOVE        ,
     IDC_RECURSEDIRS   , IDC_RECURSEDIRS   ,
     IDC_ADDDIRCOMBO   , IDC_ADDDIRCOMBO   ,
     IDOK              , IDC_ADDOK         ,
     IDCANCEL          , IDC_ADDCANCEL     ,
     0                 , 0 ,
 };

 char HelpFile[MAX_PATH];
 GetHelpFileName (HelpFile);
 HELPINFO hi = *(LPHELPINFO) lParam;

 if (hi.iContextType == HELPINFO_WINDOW &&
     hDlg != hi.hItemHandle) {
         WinHelp ((HWND) hi.hItemHandle ,
                  HelpFile ,
                  HELP_WM_HELP ,
                  (DWORD) (LPVOID) HelpID);
 }

 return TRUE;
}

/*****************************************************************************
 * Name         - DoFilenameTweaking
 * Purpose      - This function does filename tweaking for all added files.
 *                it alters the internal name and path of the file so that
 *                space processing is done and it is possibly made lower case.
 * Arguments    - File: File record to modify
 * Return Value - None
 ****************************************************************************/
void DoFilenameTweaking (AddedFile *File)
{
 WINASSERT (File);
 WINASSERT (File->InternalName);
 WINASSERT (File->InternalPath);

 // If the flag to make incoming filenames lowercase is enabled, lowercase the
 // internal name and path.
 if (Setup.LowercaseNames) {
     strlwr (File->InternalName);
     strlwr (File->InternalPath);
 }

 // Now, are spaces in filenames supposed to be converted to underscores?
 if (Setup.AddSpaceProcess == 1) {
     // Scan through both strings, converting all spaces to underscores.
     char *Temp = File->InternalName;
     while (*Temp) {
         if (*Temp == ' ')
             *Temp = '_';
         Temp++;
     }

     Temp = File->InternalPath;
     while (*Temp) {
         if (*Temp == ' ')
             *Temp = '_';
         Temp++;
     }

 } else
     // No. Should spaces be deleted?
     if (Setup.AddSpaceProcess == 2) {
         // We need to scan through both strings taking out any spaces. To
         // do that we first make a temporary buffer and copy the string into
         // it.
         char TempBuffer[MAX_PATH];
         strcpy (TempBuffer , File->InternalName);

         // make a source index and a destination index and set both to 0.
         int SrcIndex=0;
         int DstIndex=0;

         // Now, scan through the source (The temp buffer), incrementing
         // SrcIndex each time, until we get to the last character of the
         // string. For each character, if it is not a space, we put it into
         // File->InternalName[DstIndex] and update the destination index.
         while (1) {
             if (TempBuffer[SrcIndex] != ' ')
                 File->InternalName[DstIndex++] = TempBuffer[SrcIndex];

             if (TempBuffer[SrcIndex] == '\0')
                 break;
             else
                 SrcIndex++;
         }

         // Do the same for the path information
         strcpy (TempBuffer , File->InternalPath);

         SrcIndex=0;
         DstIndex=0;

         while (1) {
             if (TempBuffer[SrcIndex] != ' ')
                 File->InternalPath[DstIndex++] = TempBuffer[SrcIndex];

             if (TempBuffer[SrcIndex] == '\0')
                 break;
             else
                 SrcIndex++;
         }
     }
}

/*****************************************************************************
 * Name         - AddFileToFileList
 * Purpose      - Given a filename, a root path, and additional path info, this
 *                will add a new entry to the list of files to be added at the
 *                next save of this file. There are two situations in which
 *                this is called.
 *
 *                This can be called either to add a single file, or to add a
 *                file which is part of a directory. In the first case,
 *                AddPath (additional Path info) is NULL, and the name of the
 *                file inside the NFI file structure will be RootPath\Name.
 *                If this is called during the processesing of adding a file
 *                from a directory, then AddPath may not be NULL. In this case
 *                it gives extra path information which should be inserted
 *                between the root path and the Name. Thus the resulting name
 *                would be RootPath\AddPath\Name. This allows the code to
 *                use this same function to cleanly recurse through subdirectories,
 *                by always using the same Root Path and just updating the
 *                additional path information as it goes down the directory
 *                tree.
 * Arguments    - Name: The name of the file that is supposed to be added. This
 *                      should be a fully qualified path name to the file or it
 *                      can't be opened to be added.
 *                ChildPath: Child path information (May be NULL), which is
 *                           added between the root path and the name of the
 *                           file. this is used by the AddDirectoryToFileList
 *                           to handle sub directories during recursion.
 * Return Value - None
 ****************************************************************************/
void AddFileToFileList (char *Name , char *ChildPath)
{
 WINASSERT (Name);

 // Assume we're callign this in a loop, so pump waiting messages to keep
 // things more responsive.
 PumpMessages ();

 // This is incremented every time the function adds something. When it reaches
 // 50, the listview repains itself and this is set to 0 again. A cheap way
 // of providing feedback during long add operations.
 static int AddedItems=0;

 // If the child path information was not given, set it to a NULL string
 // so that we don't need special cases later.
 if (ChildPath == NULL)
     ChildPath = "";

 // Make a node
 AddedFile *File = (AddedFile *) new char[sizeof (AddedFile)];
 memset (File , 0 , sizeof (AddedFile));

 // Store the full name of the file in the AddFrom member of the node, so
 // that later on we can tell what file to open when we want to add it.
 File->AddFrom = new char[strlen (Name) + 1];
 strcpy (File->AddFrom , Name);

 // Now, determine what the path information is goign to be. To start, we
 // want to strip any path information off of the given filename. We then
 // use a temporary buffer to construct the filename and the path informations.
 static char FName[_MAX_FNAME];
 static char FExt[_MAX_EXT];
 static char Temp[_MAX_FNAME + _MAX_EXT];
 _splitpath (Name , NULL , NULL , FName , FExt);
 strcpy (Temp , FName);
 strcat (Temp , FExt);

 static char FPath[MAX_PATH];

 // If the combined lenth of the root path, the additional path, and the
 // filename is too big, then we make the root path be just "/" and remove
 // additional path info. THis saves processing later and shows people to
 // watch out using incredibly long names.
 if (strlen (Setup.AddDir) + strlen (ChildPath) + 2 + strlen (Temp) > MAX_PATH) {
     strcpy (FPath , "/");
 } else {
     strcpy (FPath , Setup.AddDir);
     strcat (FPath , ChildPath);

     // We need a trailing slash if there was additional path information
     if (strlen (ChildPath) != 0)
         strcat (FPath , "/");
 }

 // Now we know the filename and path of this file. Update the node.
 File->InternalName = new char[strlen (Temp) + 1];
 strcpy (File->InternalName , Temp);
 File->InternalPath = new char[strlen (FPath) + 1];
 strcpy (File->InternalPath , FPath);

 // Tweak the internal names if we have to.
 DoFilenameTweaking (File);

 // Now add this to the linked list. All additions are at the head. If there
 // isn't one, this is the head. Otherwise, this becomes the new head.
 if (AddedFileList == NULL) {
     AddedFileList = File;
 } else {
     // The next of this node points to what the head is now
     File->Next = AddedFileList;
     // The Prev of the head node points at us
     AddedFileList->Prev = File;
     // The head of the list is now us
     AddedFileList = File;
 }

 // Update the count of files that are going to be added.
 FilesToAdd++;

 // Insert an entry for this file into the listview. Note that normally we
 // give an lParam for any entry that points to the struct which contains the
 // information for it. Since we don't have that yet, we instead insert a
 // pointer to the added File linked list entry which referece this file. The
 // code can determine what sort of node lParam points to by examing the
 // image index of the entry. An index of 1 means an added file, while 0
 // means a loaded file. With this, the program can allow changes to any
 // entered file data with impunity.
 static LV_ITEM lvi;
 CResString Unknown (hInst , IDS_UNKNOWNVALUE);

 lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;

 lvi.iItem    = 0;
 lvi.iSubItem = 0;
 lvi.pszText  = File->InternalName;
 lvi.iImage   = 1;
 lvi.lParam   = (LPARAM) File;

 // Insert the item
 ListView_InsertItem  (hwndListview , &lvi);

 // Set the size text. Try to get the file size from disk. If at any time it fails
 // we let it pass and just use the unknown string for the size. Otherwise we add
 // the real one.
 HANDLE TempHandle = CreateFile (Name , GENERIC_READ , FILE_SHARE_READ , NULL ,
                        OPEN_EXISTING , 0 , NULL);
 if (TempHandle != INVALID_HANDLE_VALUE) {
     DWORD Size = GetFileSize (TempHandle , NULL);
     CloseHandle (TempHandle);

     static char Buf[128];
     if (Size != 0xFFFFFFFF) {
         ultoa (Size , Buf , 10);
     } else
         strcpy (Buf , Unknown);
     
     ListView_SetItemText (hwndListview , 0 , 1 , Buf);
 } else 
     ListView_SetItemText (hwndListview , 0 , 1 , Unknown);

 // Set the offset text
 ListView_SetItemText (hwndListview , 0 , 2 , Unknown);

 // Set the path
 ListView_SetItemText (hwndListview , 0 , 3 , File->InternalPath);

 // Try to add this path to the binary tree of path information.
 NFIPathInfo = AddPathToTree (NFIPathInfo , FPath);

 // Do a little update. Every 50 items make the listview repaint itself so
 // that the user gets some feedback about what's going on
 AddedItems++;
 if (AddedItems == 50) {
     UpdateWindow (hwndListview);
     AddedItems = 0;
 }
}

/*****************************************************************************
 * Name         - AddDirectoryToFileList
 * Purpose      - This adds a directory to a file list. It does this by doing
 *                a scan for all files which are in a given directory and
 *                calling the AddFileToFileList function to add it. The
 *                function handles recursion as well if needed by calling
 *                itself again if it comes across a directory name and it
 *                has been asked to recurse through directories.
 * Arguments    - Start: This is the main root directory in which the file
 *                       add operation started. This never changes but is
 *                       needed for recursion. It always has a trailing slash
 *                Directory: This is the directory which is currently being
 *                           scanned. It is also a compelte path name with a
 *                           trailing slash
 * Return Value - None
 ****************************************************************************/
void AddDirectoryToFileList (char *Start , char *Directory)
{
 WINASSERT (Start);
 WINASSERT (Directory);

 // We need to find a search path. This is easy. It is Directory\*.*.
 char SearchPath[MAX_PATH];
 strcpy (SearchPath , Directory);
 strcat (SearchPath , "*.*");

 // Now, if Start == Directory, then any files added always go into
 // the NFI directory indicated by Setup.AddDir. If they are not equal, we
 // need to strip the parent information from the given directory to end up
 // with only the child portion. That is then used to indicate that we are
 // adding to a child of Setup.AddDir.
 char *AdditionalDirInfo = NULL;
 if (strcmpi (Start , Directory) != 0) {
     // Since Directory has an inital substring of Start, all we have to do
     // is find the length of Start and use that as an index into Directory
     // to get the child information.
     AdditionalDirInfo = new char[strlen (Directory) - strlen (Start) + 2];
     strcpy (AdditionalDirInfo , Directory + strlen (Start));

     // Removing the trailing slash
     AdditionalDirInfo[strlen (AdditionalDirInfo) - 1] = '\0';

     // Convert all slashes
     char *Temp = AdditionalDirInfo;
     while (*Temp) {
         if (*Temp == '\\')
             *Temp = '/';
         Temp++;
     }
 }

 HANDLE          SearchHandle;
 WIN32_FIND_DATA wfd;
 BOOL            Done=FALSE;

 // Start searching
 SearchHandle = FindFirstFile (SearchPath , &wfd);
 if (SearchHandle == INVALID_HANDLE_VALUE) {
     if (AdditionalDirInfo) delete[] AdditionalDirInfo;
     return;
 }

 // Continue until we get all the files.
 while (!Done) {
     // Is this a directory?
     if ((wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0) {
         // No, it is a file. We make a string that contains the complete
         // pathname of this file, then add that
         char FileBeingAdded[MAX_PATH];

         // The full path of this file is the directory we're scanning plus
         // the name of the file itself
         strcpy (FileBeingAdded , Directory);
         strcat (FileBeingAdded , wfd.cFileName);

         AddFileToFileList (FileBeingAdded , AdditionalDirInfo);
     } else {
         // It's a dir? Are we supposed to recurse?
         if (Setup.RecurseDirs) {
             // Is it a valid directory and not . or ..?
             if (strcmp (wfd.cFileName , ".") != 0 && strcmp (wfd.cFileName , "..") != 0) {
                 // Is it not a system directory?
                 if ((wfd.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM) == 0) {
                     // Make a new search directory. This is the one we're
                     // searching plus the one we just found.
                     char NewChild[MAX_PATH];
                     strcpy (NewChild , Directory);
                     strcat (NewChild , wfd.cFileName);
                     strcat (NewChild , "\\");
                     AddDirectoryToFileList (Start , NewChild);
                 }
             }
         }
     }

     // Find the next file.
     Done = !FindNextFile (SearchHandle , &wfd);
 }

 FindClose (SearchHandle);
 if (AdditionalDirInfo) delete[] AdditionalDirInfo;
}

/*****************************************************************************
 * Name         - UpdateAddedFileList
 * Purpose      - This adds the files the user has selected to the linked list
 *                which stores all of the files waiting to be added to the
 *                NFI file at the next save. For files, it adds a new node.
 *                For directories it will add all the files in that directory
 *                recursing if it was told to. The number of files that where
 *                added is updated as well.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void UpdateAddedFileList (void)
{
 // Find out how many listbox entries there are, and loop one for each of
 // them.
 int Count = SendMessage (hwndAddListBox , LB_GETCOUNT , 0 , 0);

 // If there is nothing to add, return
 if (Count == 0)
     return;

 char ThisEntry[MAX_PATH];
 for (int Index = 0 ; Index < Count ; Index++) {
     // Get this entry
     SendMessage (hwndAddListBox , LB_GETTEXT , (WPARAM) Index , (LPARAM) ThisEntry);

     // If this entry is a filename, just add it. Otherwise we use a helper
     // function to add all the files in a directory. We tell if an item is
     // a filename or not by wether it has a trailing slash on it.
     if (ThisEntry[strlen (ThisEntry) - 1] == '\\') {
         // Entry is a directory
         AddDirectoryToFileList (ThisEntry , ThisEntry);
     } else {
         // Entry is a file. Add it
         AddFileToFileList (ThisEntry , NULL);
     }
 }

 // Make the listview update and then update the statusbar with the new info.
 UpdateWindow (hwndListview);
 UpdateFileCount ();
 SetFileStatus (TRUE);
}

/*****************************************************************************
 * Name         - CAddDialogDispatcher::CmdDone
 * Purpose      - This function handles command messages for the button and
 *                the standard windows close controls.
 * Arguments    - hDlg: Handle of the dialog box
 *                wCommand: Command ID of the button that was pressed
 * Return Value - TRUE
 ****************************************************************************/
LRESULT CAddDialogDispatcher::CmdDone (HWND hDlg, WORD wCommand,
                                       WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 // Save the state of the recurse directories checkbox into the registry for
 // next time. Also, we grab the directory the user wants the files added to
 // and save that as well.
 if (wCommand == IDOK) {
     Setup.RecurseDirs = (IsDlgButtonChecked (hDlg , IDC_RECURSEDIRS) == BST_CHECKED);
     GetWindowText (GetDlgItem (hDlg , IDC_ADDDIRCOMBO) ,
                    Setup.AddDir ,
                    MAX_PATH);

     // If the string is empty, make it the root string
     if (Setup.AddDir[0] == '\0')
         strcpy (Setup.AddDir , "/");

     // Convert slashes
     char *Temp = Setup.AddDir;
     while (*Temp) {
         if (*Temp == '\\')
             *Temp = '/';
         Temp++;
     }

     // Make sure it has a trailing slash and a leading slash
     if (Setup.AddDir[0] != '/') {
         char Temp[MAX_PATH] = "/";
         strcat (Temp , Setup.AddDir);
         strncpy (Setup.AddDir , Temp , MAX_PATH);
         Setup.AddDir[MAX_PATH - 2] = '\0';
     }
     if (Setup.AddDir[strlen (Setup.AddDir) - 1] != '/')
         strcat (Setup.AddDir , "/");

     NFIPathInfo = AddPathToTree (NFIPathInfo , Setup.AddDir);
     SaveSetupStructToRegistry (Setup);

     // Set up the global listbox variable and then set it's parent to the
     // main window so that it doesn't get destroyed.
     hwndAddListBox = GetDlgItem (hDlg , IDC_ADDLISTBOX);
     SetParent (hwndAddListBox , hwndMain);
     ShowWindow (hwndAddListBox , SW_HIDE);
 }

 // End the dialog
 EndDialog (hDlg , wCommand);
 return TRUE;
}

/*****************************************************************************
 * Name         - ShellBrowserCallback
 * Purpose      - This is a callback function used by the directory broswer
 *                dialog. It is given messages that would let us track what
 *                the dialog is doing. Instead, we use it to force the selected
 *                dir to be the same one that was selected last time, since
 *                there is no easy way to do so.
 * Arguments    - hWnd: Handle of the browse dialog box
 *                uMessage: Message it is giving us
 * Return Value - 0
 ****************************************************************************/
static int CALLBACK ShellBrowserCallback (HWND hWnd , UINT uMessage ,
                                          LPARAM /*lParam*/ , LPARAM /*lpData*/)
{
 // Is the dialog telling us that it has been initialized, AND do we have
 // a previously selected directory?
 if (uMessage == BFFM_INITIALIZED && SelectedDirToAdd[0] != '\0') {
     // Yes. Tell it to select the directory which is identified by the
     // string that the last call to the directory browser gave us. This
     // eases the user's burden of adding more than one directory from the
     // same relative position. Nice for people like me who have drives out
     // the wazoo
     SendMessage (hWnd , BFFM_SETSELECTION ,
                  (WPARAM) TRUE , (LPARAM) SelectedDirToAdd);
 }

 return 0;
}

/*****************************************************************************
 * Name         - CAddDialogDispatcher::CmdAddFile
 * Purpose      - This functions handles command messages for the IDB_ADDFILE
 *                button
 * Arguments    - hDlg: Handle of the dialog box
 * Return Value - TRUE
 ****************************************************************************/
LRESULT CAddDialogDispatcher::CmdAddFile (HWND hDlg, WORD /*wCommand*/,
                                          WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 OPENFILENAME of;
 CResString   Title  (hInst , IDS_ADDFILETITLE);
 CResString   Filter (hInst , IDS_ALLFILEFILTER);

 // Since you cannot use the \0 notation in a resource string (it stops loading
 // there), we specify the filter using | marks intead of null characters.
 // Here we get a temporary pointer to the string data and scan through it
 // converting all of the | characters to null ones.
 char *Temp = Filter;
 while (*Temp) {
     if (*Temp == '|')
         *Temp = '\0';
     Temp++;
 }

 // Make the buffer that we use to store the selected filenames
 char *FilenameBuffer = new char[ADD_FILENAME_BUFFER_SIZE];
 FilenameBuffer[0] = '\0';

 // Initialize the struct that is used in the open file common dialog box
 of.lStructSize       = sizeof (of);
 of.hwndOwner         = hDlg;
 of.hInstance         = hInst;
 of.lpstrFilter       = Filter;
 of.lpstrCustomFilter = NULL;
 of.nFilterIndex      = 0;
 of.lpstrFile         = FilenameBuffer;
 of.nMaxFile          = ADD_FILENAME_BUFFER_SIZE;
 of.lpstrFileTitle    = NULL;
 of.lpstrInitialDir   = NULL;
 of.lpstrTitle        = Title;
 of.lpstrDefExt       = NULL;
 of.Flags             = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY |
                        OFN_PATHMUSTEXIST | OFN_ALLOWMULTISELECT |
                        OFN_EXPLORER;

 // Allow the user to select files. If the function returns 0 there was either
 // an error or else the user clicked cancel. Either way, get rid of our
 // buffer and quit
 if (!GetOpenFileName (&of)) {
     delete[] FilenameBuffer;
     return TRUE;
 }

 // We have a multi select explorer style dialog box. That means that the
 // returned string is a series of null terminated strings, the last of which
 // is double null terminated. The first is the directory that the files are
 // in, and the rest are the actual files. We first get the directory out,
 // and then advance through the string getting all the files. For each file
 // we append it to a copy of the directory. Note that to get to the next
 // string we use an index that starts at 0, and is increased by the length
 // of the string plus one. When we end up at a string that is 0 length
 // we know we're done.
 int Index = 0;
 char Directory[MAX_PATH];
 char Filename[MAX_PATH];

 // Get the directory
 strcpy (Directory , FilenameBuffer);

 // Determine the index the next string starts at
 Index = strlen (FilenameBuffer) + 1;

 // Maintain a count of added files
 int Count = 0;

 // Cache the handle to the listbox
 HWND Listbox = GetDlgItem (hDlg , IDC_ADDLISTBOX);

 // Loop while we have files
 while (strlen (FilenameBuffer + Index) != 0) {
     // Copy the directory to the file buffer, then copy the filename, then
     // add the string to the listbox.
     strcpy (Filename , Directory);
     strcat (Filename , "\\");
     strcat (Filename , FilenameBuffer + Index);
     SendMessage (Listbox , LB_INSERTSTRING ,
                  (WPARAM) -1 , (LPARAM) Filename);

     // Advance the index to the next filename
     Index += strlen (FilenameBuffer + Index) + 1;

     // Increment the count of added files
     Count++;
 }

 // If we get here and the count is 0, then only one file was selected. In
 // that case what we think is the directory is actually the file, so add
 // that
 if (Count == 0)
     SendMessage (Listbox , LB_INSERTSTRING ,
                  (WPARAM) -1 , (LPARAM) Directory);

 // Get rid of the filename buffer
 delete[] FilenameBuffer;
 return TRUE;
}

/*****************************************************************************
 * Name         - CAddDialogDispatcher::CmdAddDir
 * Purpose      - This function handles command messages for the IDB_ADDDIR
 *                button
 * Arguments    - hDlg: Handle of the dialog box
 * Return Value - TRUE
 ****************************************************************************/
LRESULT CAddDialogDispatcher::CmdAddDir (HWND hDlg, WORD /*wCommand*/,
                                         WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 BROWSEINFO bi;
 CResString Help (hInst , IDS_ADDDIRHELP);

 // Set up the browse info struct. This tells the shell directory browser
 // dialog box how to add
 bi.hwndOwner = hDlg;
 bi.pidlRoot = NULL;
 bi.pszDisplayName = SelectedDirToAdd;
 bi.lpszTitle = Help;
 bi.ulFlags = BIF_RETURNONLYFSDIRS;
 bi.lpfn = (BFFCALLBACK) ShellBrowserCallback;

 // Let the user select a folder
 LPITEMIDLIST pidl = SHBrowseForFolder (&bi);

 // If nothing was selected, return
 if (!pidl)
     return TRUE;

 // See if we can manage to get directory out of the identifier list. Note
 // that I don't think is is actually needed, but all the examples I've
 // ever seen do it.
 BOOL bSuccess = SHGetPathFromIDList (pidl , SelectedDirToAdd);

 // Try to get a handle to the allocator the shell uses. If it doesn' work,
 // toss an error.
 IMalloc* pMalloc;
 if (SHGetMalloc (&pMalloc) != NOERROR)
     resThrow (hInst , IDS_EXCEPTMALLOC);

 // Release the id list and then release the allocator
 pMalloc->Free (pidl);
 pMalloc->Release();

 // If we actually got a folder, then add it to the listbox
 if (bSuccess) {
     // We make a copy of the folder name before we do anything, because
     // SelectedDirToAdd must always remain the same. If we add the slash to
     // it, then the next time the directory browser is invoked, it can't
     // find the directory. A dumb limitation of microsoft. To get around
     // that, we make a copy of the directory and fiddle with that.
     char Folder[MAX_PATH];
     strcpy (Folder , SelectedDirToAdd);

     // Put a trailing slash if there isn't one
     if (Folder[strlen (Folder) - 1] != '\\')
         strcat (Folder , "\\");

     SendMessage (GetDlgItem (hDlg , IDC_ADDLISTBOX) ,
                  LB_INSERTSTRING ,
                  -1 , (LPARAM) Folder);
 }

 return TRUE;
}

/*****************************************************************************
 * Name         - CAddDialogDispatcher::CmdRemove
 * Purpose      - This function handles command messages for the IDB_REMOVE
 *                button
 * Arguments    - hDlg: Handle of the dialog box
 * Return Value - TRUE
 ****************************************************************************/
LRESULT CAddDialogDispatcher::CmdRemove (HWND hDlg, WORD /*wCommand*/,
                                         WORD /*wNotify*/, HWND /*hwndCtrl*/)
{
 // Get the handle to the list box
 HWND Listbox = GetDlgItem (hDlg , IDC_ADDLISTBOX);

 // Find out how many selected items there are
 int SelectCount = SendMessage (Listbox , LB_GETSELCOUNT , 0 , 0);

 // If there aren't any selected items, return
 if (SelectCount == 0)
     return TRUE;

 // Make a buffer to store the selected items in
 int *SelectedItems = new int[SelectCount];

 // Tells the listbox to give us the selected list of items
 SendMessage (Listbox , LB_GETSELITEMS , (WPARAM) SelectCount , (LPARAM) SelectedItems);

 // Loop deleting every selected item from the list box. Note, we start from
 // the last index and move toward the start. This is because if we delete
 // item 0, the count of everything else is altered
 for (int Index = SelectCount - 1 ; Index >= 0 ; Index--)
     SendMessage (Listbox , LB_DELETESTRING , SelectedItems[Index] , 0);

 // Get rid of our buffer
 delete[] SelectedItems;

 return TRUE;
}

/****************************************************************************/

