/****************************************************************************/

#include "nfi.h"

// This defines the maximum amount of data that the read or write code used
// in the save operation will write in one time, either when copying data
// over to a new file or adding one in. Making this bigger makes the program
// require more memory but if yuo anticipate alot of large files it may be
// worth it.
#define MAX_FILE_BUFFER 128000

/*****************************************************************************
 * Name         - DestroyAddedFileList
 * Purpose      - This function compeltely destroys the list of files waiting
 *                to be added. This is used whenever a new file is created
 *                or a save operation completes successfully.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DestroyAddedFileList (void)
{
 // Nukify the linked list of files to be added.
 AddedFile *File = AddedFileList;
 while (File) {
     delete[] File->AddFrom;
     delete[] File->InternalName;
     delete[] File->InternalPath;

     // Get a char pointer to this node (We allocated the memory as an
     // array of chars.
     char *Temp = (char *) File;

     // Advance to the next entry before we delete this one
     File = File->Next;

     delete[] Temp;
 }
 AddedFileList = NULL;
}

/*****************************************************************************
 * Name         - WriteErrorAndReset
 * Purpose      - This is a helper function called by the DoSave function.
 *                Whenever there is an error writing or opening the file, this
 *                is called to display a message box, and then closes the
 *                given file.
 * Arguments    - File: Handle to the temporary file
 *                TitleId: Res ID of the string used for the error title
 *                ErrorId: Res ID of the string used for the error message
 *                FName: Additional string paramter, NULL means don't worry.
 * Return Value - None
 ****************************************************************************/
void WriteErrorAndReset (HANDLE File , int TitleId , int ErrorId , char *FName=NULL)
{
 static char Buffer[2*MAX_PATH];

 // Display the error message
 CResString Title (hInst , TitleId);
 CResString Msg   (hInst , ErrorId);

 // If there was a filename passed in, wsprintf it into the loaded message
 // string. Otherwise, just copy the string over to the buffer.
 if (FName)
     wsprintf (Buffer , Msg , FName);
 else
     strcpy (Buffer , Msg);

 MessageBox (hwndMain , Buffer , Title , MB_OK | MB_ICONERROR);

 CloseHandle (File);
}

/*****************************************************************************
 * Name         - WriteNFIHeader
 * Purpose      - This function writes the given NFI Header to the given file.
 *                If there is any error the function cleans everything up and
 *                returns false. Otherwise it returns true.
 * Arguments    - File: Handle to the temporary file
 *                Header: The header to output to the file
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL WriteNFIHeader (HANDLE File , NFIHeader &Header)
{
 DWORD BytesWritten;

 // Output each part of the header to the file. On any error, call the error
 // display and reset function and return false meaning error
 if (!WriteFile (File , Header.IDTag ,
                sizeof (Header.IDTag) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Header.IDTag)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Header.Version) ,
                sizeof (Header.Version) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Header.Version)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Header.Reserved) ,
                sizeof (Header.Reserved) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Header.Reserved)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Header.DirOffset) ,
                sizeof (Header.DirOffset) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Header.DirOffset)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }

 // If we get here it worked so return TRUE.
 return TRUE;
}

/*****************************************************************************
 * Name         - WriteDirHeader
 * Purpose      - This function writes the given directory header to the given
 *                file. If there is any error the function cleans everything
 *                up and returns false. Otherwise it returns true.
 * Arguments    - File: Handle to the temporary file
 *                Header: The header to output to the file
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL WriteDirHeader (HANDLE File , NFIDirHeader &Header)
{
 DWORD BytesWritten;

 // Output each part of the header to the file. On any error, call the error
 // display and reset function and return false meaning error
 if (!WriteFile (File , &(Header.NumEntries) ,
                sizeof (Header.NumEntries) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Header.NumEntries)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Header.Reserved) ,
                sizeof (Header.Reserved) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Header.Reserved)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }

 // If we get here it worked so return TRUE.
 return TRUE;
}

/*****************************************************************************
 * Name         - WriteDirectoryEntry
 * Purpose      - This function writes the given directory header to the given
 *                file. If there is any error the function cleans everything
 *                up and returns false. Otherwise it returns true.
 * Arguments    - File: Handle to the temporary file
 *                Header: The header to output to the file
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL WriteDirectoryEntry (HANDLE File , NFIDirEntry *Entry)
{
 DWORD BytesWritten;

 // Output each part of the header to the file. On any error, call the error
 // display and reset function and return false meaning error
 if (!WriteFile (File , Entry->Name ,
                sizeof (Entry->Name) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Entry->Name)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Entry->Offset) ,
                sizeof (Entry->Offset) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Entry->Offset)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Entry->Length) ,
                sizeof (Entry->Length) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Entry->Length)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Entry->Reserved[0]) ,
                sizeof (Entry->Reserved[0]) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Entry->Reserved[0])) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }
 if (!WriteFile (File , &(Entry->Reserved[1]) ,
                sizeof (Entry->Reserved[1]) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (Entry->Reserved[1])) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORWRITING);
     return FALSE;
 }

 // If we get here it worked so return TRUE.
 return TRUE;
}

/*****************************************************************************
 * Name         - MakeRealEntry
 * Purpose      - This is used during save operations to create a real
 *                directrory entry from the information in the nodes that
 *                represent a file that is supposed to be added.
 * Arguments    - File: Pointer to the added file node to make the new entry
 *                      from.
 *                Entry: The struct to fill with the proper information
 * Return Value - Pointer to the new directory entry
 ****************************************************************************/
void MakeRealEntry (AddedFile *File , NFIDirEntry &Entry)
{
 // Now, fill in the relevant fields. Start by zeroing out everything.
 memset (&Entry , 0 , sizeof (NFIDirEntry));

 // All we need to do is make a name. The length and offset will be added
 // later on when we actually know it.
 strcpy (Entry.Name , File->InternalPath);
 strcat (Entry.Name , File->InternalName);
}

/*****************************************************************************
 * Name         - CopyDataChunk
 * Purpose      - This function, used during a save operation, does the work
 *                of copying a chunk of data from one file to another. On
 *                Error, the file represented by "To" is the one which is
 *                closed. The from file is not altered.
 * Arguments    - From: File to get the data from
 *                To: File to write the data to
 *                Size: Number of bytes to transfer
 *                TitleID: String ID of the title message to use
 *                ErrorID: String ID of the error message to use
 *                FName: Additional error paramter
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL CopyDataChunk (HANDLE From , HANDLE To , int Size , int TitleID ,
                    int ErrorID , char *FName)
{
 static char Buffer[MAX_FILE_BUFFER];
 static DWORD BytesRead,BytesWritten;

 // First, try to read the data from the "From" file.
 if (!ReadFile (From , Buffer , Size  , &BytesRead , NULL) ||
                BytesRead != Size) {
     WriteErrorAndReset (To , TitleID , ErrorID , FName);
     return FALSE;
 }

 // Now, try to write the data to the "To" file.
 if (!WriteFile (To , Buffer , Size , &BytesWritten , NULL) ||
                BytesWritten != Size) {
     WriteErrorAndReset (To , TitleID , ErrorID , FName);
     return FALSE;
 }

 // We assume that we are being called in a loop, so pump any waiting messages
 // to keep things fairly responsive.
 PumpMessages ();

 // If we get here the operation was a resounding success.
 return TRUE;
}

/*****************************************************************************
 * Name         - CopyNFIItem
 * Purpose      - This function does the work of copying an NFI item from the
 *                original NFI file to the new file. Once that is done, if
 *                successfull, the function seeks back to update the directory
 *                entry for this file, then returns to the end of the file and
 *                returns TRUE. If any error occurs, the function will close
 *                the passed in file, display an error message and then return
 *                FALSE.
 * Arguments    - File: File to write to
 *                lvi: Struct specifying the list view item we need to do the
 *                     job.
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL CopyNFIItem (HANDLE File , LV_ITEM &lvi)
{
 DWORD BytesWritten;

 // File represents the new NFI file we are creating. We are wanting to copy
 // the entry represented by the lParam of the listview item we where given.
 // First thing to do is to get a handle to that entry.
 NFIDirEntry *Entry = (NFIDirEntry *) lvi.lParam;

 // Now, the entry represents the information as it was loaded from the main
 // NFI file. So we seek to that file position in the Main NFI File.
 SetFilePointer (MainNFIFile , Entry->Offset , NULL , FILE_BEGIN);

 // Determine where in the new NFI file this entry is. We can determine that
 // be querying the file's length, which should tell us what the last byte
 // of the file is, which is the information that we want.
 unsigned long ThisOffset = GetFileSize (File , NULL);

 // Now, copy all of the data from that file to the new file. First, determine
 // how many bytes we have to read.
 LONG BytesToRead = Entry->Length;
 LONG BytesThisLoop;

 // Loop until all the bytes have been read and copied over.
 while (BytesToRead) {
     // Determine how many bytes we should go this chunk.
     if (BytesToRead > MAX_FILE_BUFFER)
         BytesThisLoop = MAX_FILE_BUFFER;
     else
         BytesThisLoop = BytesToRead;

     // Copy the data. Return on error, the new file will already have been
     // closed.
     if (!CopyDataChunk (MainNFIFile , File , BytesThisLoop , IDS_ERRORWRITING , IDS_ERRORCOPYING , NULL))
         return FALSE;

     // Update the bytes left to read.
     BytesToRead -= BytesThisLoop;
 }

 // Once we get here, we have successfully written to the file. Now we need to
 // go back and update the file offset. Note that the other information about
 // this file, which is it's length, was correctly written to the new file
 // when the directory was created, because this file existed in the original
 // NFI File. So, we need to determine where in the new file to write the
 // offset. This is simple. As it turns out, what we must do is add up the
 // number of bytes for the  main header, the directory header, plus all of
 // the directory entries that came before this one. That will give us the
 // exact file location of the start of this file's directory entry. We then
 // add in the length of the name of this file to skip past that field. That
 // will put us on top of the Offset field. We can then output our offset.
 // We use the iItem member of the passed listview item struct to determine
 // what file number we are.

 // Determine the number of bytes used for the file header.
 LONG SeekPos = (strlen (NFI_HEADERID) + 1) + sizeof (unsigned char) +
                (2 * sizeof (unsigned long));

 // Add to that the number of bytes used for the directory header.
 SeekPos += (2 * sizeof (unsigned long));

 // Determine the number of bytes used to hold a directory entry. We can't
 // just use sizeof because there are members in the entry that are never
 // written to disk.
 LONG EntrySize = NFI_NAMELEN + (4 * sizeof (unsigned long));

 // Adjust the Seek position to take the item index into account. If we are
 // item N then N-1 items came before us. However, the item index we have is
 // already 0 based so we don't have to worry about that. We also add the size
 // of the name field of entry to skip past the name.
 SeekPos += (EntrySize * lvi.iItem) + NFI_NAMELEN;

 // TaDa, now seek to the file position we have all worked out and output the
 // offset. Once that is done, seek back to the end of the file.
 SetFilePointer (File , SeekPos , NULL , FILE_BEGIN);
 if (!WriteFile (File , &ThisOffset , sizeof (ThisOffset) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (ThisOffset)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORCOPYING , NULL);
     return FALSE;
 }
 SetFilePointer (File , 0 , NULL , FILE_END);

 // If we get here it worked.
 return TRUE;
}

/*****************************************************************************
 * Name         - AddFileToNFI
 * Purpose      - This function is the counterpart of the CopyNFIItem function
 *                except that it adds a file from the local file system to the
 *                NFI, rather than copying over from the original file. In all
 *                other respects, the function performs the same tasks.
 * Arguments    - File: File to write to
 *                lvi: Struct specifying the list view item we need to do the
 *                     job.
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL AddFileToNFI (HANDLE File , LV_ITEM &lvi)
{
 DWORD BytesWritten;

 // File represents the new NFI file we are creating. We are wanting to add
 // a new file. The name of that file is stored in the lParam of the passed
 // in listview item. First thing to do is to get a handle to that entry
 AddedFile *Entry = (AddedFile *) lvi.lParam;

 // Attempt to open the file we want to add. If there is an error, close
 // the file we where given and return false after displaying an error
 // message.
 HANDLE FileBeingAdded = CreateFile (Entry->AddFrom , GENERIC_READ ,
                                     FILE_SHARE_READ , NULL , OPEN_EXISTING ,
                                     FILE_FLAG_RANDOM_ACCESS , NULL);
 if (FileBeingAdded == INVALID_HANDLE_VALUE) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORADDING , Entry->AddFrom);
     return FALSE;
 }

 // We opened the file alright. What we need to do now is determine how long
 // it is.

 unsigned long ThisLength = GetFileSize (FileBeingAdded , NULL);

 // Determine where in the new NFI file this entry is. We can determine that
 // be querying the file's length, which should tell us what the last byte
 // of the file is, which is the information that we want.
 unsigned long ThisOffset = GetFileSize (File , NULL);

 // Now, copy all of the data from that file to the new file. First, determine
 // how many bytes we have to read.
 LONG BytesToRead = ThisLength;
 LONG BytesThisLoop;

 // Loop until all the bytes have been read and copied over.
 while (BytesToRead) {
     // Determine how many bytes we should go this chunk.
     if (BytesToRead > MAX_FILE_BUFFER)
         BytesThisLoop = MAX_FILE_BUFFER;
     else
         BytesThisLoop = BytesToRead;

     // Copy the data. Return on error, the new file will already have been
     // closed.
     if (!CopyDataChunk (FileBeingAdded , File , BytesThisLoop , IDS_ERRORWRITING , IDS_ERRORADDING , Entry->AddFrom)) {
         CloseHandle (FileBeingAdded);
         return FALSE;
     }

     // Update the bytes left to read.
     BytesToRead -= BytesThisLoop;
 }

 // Close the file we added, we're done with it.
 CloseHandle (FileBeingAdded);

 // Once we get here, we have successfully written to the file. Now we need to
 // go back and update the file offset and legnth. So, we need to determine
 // where in the new file to write the data. This is simple. As it turns out,
 // what we must do is add up the number of bytes for the  main header, the
 // directory header, plus all of the directory entries that came before this
 // one. That will give us the exact file location of the start of this file's
 // directory entry. We then add in the length of the name of this file to
 // skip past that field. That will put us on top of the Offset field. We can
 // then output our offset. We use the iItem member of the passed listview
 // item struct to determine what file number we are.

 // Determine the number of bytes used for the file header.
 LONG SeekPos = (strlen (NFI_HEADERID) + 1) + sizeof (unsigned char) +
                (2 * sizeof (unsigned long));

 // Add to that the number of bytes used for the directory header.
 SeekPos += (2 * sizeof (unsigned long));

 // Determine the number of bytes used to hold a directory entry. We can't
 // just use sizeof because there are members in the entry that are never
 // written to disk.
 LONG EntrySize = NFI_NAMELEN + (4 * sizeof (unsigned long));

 // Adjust the Seek position to take the item index into account. If we are
 // item N then N-1 items came before us. However, the item index we have is
 // already 0 based so we don't have to worry about that. We also add the size
 // of the name field of entry to skip past the name.
 SeekPos += (EntrySize * lvi.iItem) + NFI_NAMELEN;

 // TaDa, now seek to the file position we have all worked out and output the
 // offset and length. Once that is done, seek back to the end of the file.
 SetFilePointer (File , SeekPos , NULL , FILE_BEGIN);
 if (!WriteFile (File , &ThisOffset , sizeof (ThisOffset) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (ThisOffset)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORADDING , Entry->AddFrom);
     return FALSE;
 }
 if (!WriteFile (File , &ThisLength , sizeof (ThisLength) , &BytesWritten , NULL) ||
                BytesWritten != sizeof (ThisLength)) {
     WriteErrorAndReset (File , IDS_ERRORWRITETITLE , IDS_ERRORADDING , Entry->AddFrom);
     return FALSE;
 }
 SetFilePointer (File , 0 , NULL , FILE_END);

 // If we get here it worked.
 return TRUE;
}

/*****************************************************************************
 * Name         - DoSave
 * Purpose      - This performs the actual save operation. This code assumes
 *                that the Filename member is valid. It constructs a new NFI
 *                file which adds to or deletes from the file that is
 *                currently loaded, or just creates a new one if there was
 *                not previously one loaded.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DoSave (void)
{
 // If there isn't a filename at the moment, then we need to do a save as and
 // not a save. The SaveAs code will set filename and then call DoSave again.
 if (Filename[0] == '\0') {
     DoSaveAs ();
     return;
 }

 // First thing we want to do is take the full name of the filename we are
 // supposed to save to, split the path off of it and make a new, temporary
 // name out of it. We save by creating a temporary file, and once that is
 // done, we rename it to the name we where asked to save. That way, if there
 // is already a file loaded and we are saving it again, we don't ruin the
 // file if there is an error of some kind.
 char TempDrive[_MAX_DRIVE];
 char TempPath[_MAX_PATH];
 char TempFName[_MAX_PATH];

 _splitpath (Filename , TempDrive , TempPath , NULL , NULL);
 strcpy (TempFName , TempDrive);
 strcat (TempFName , TempPath);
 strcat (TempFName , "$$$$$$$$.$$$");

 // Try to create a brand new file with this temp name. We will write to that
 // file.
 HANDLE NewFile = CreateFile (TempFName , GENERIC_WRITE , 0 , NULL ,
                              CREATE_ALWAYS ,
                              FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS ,
                              NULL);

 if (NewFile == INVALID_HANDLE_VALUE) {
     CResString Title   (hInst , IDS_ERRORWRITETITLE);
     CResString Message (hInst , IDS_ERRORCREATING);

     MessageBox (hwndMain , Message , Title , MB_OK | MB_ICONERROR);
     return;
 }

 // Now, we have to make a header and a directory header for this file and
 // write those out.
 NFIHeader    Header;
 NFIDirHeader DirHeader;

 // Make the header
 memset (&Header , 0 , sizeof (Header));
 strcpy (Header.IDTag , NFI_HEADERID);
 Header.Version = NFI_VERSION;
 Header.DirOffset = sizeof (Header.IDTag) + sizeof (Header.Version) +
                    sizeof (Header.Reserved) + sizeof (Header.DirOffset);

 memset (&DirHeader , 0 , sizeof (DirHeader));
 DirHeader.NumEntries = FilesInNFI + FilesToAdd;

 // The file was created, so create the dialog box and get things rolling.
 WINASSERT (hwndProgressDlg == NULL);
 CreateDialogParam (hInst, MAKEINTRESOURCE (IDD_PROGRESSDLG) , hwndMain , (DLGPROC) ProgressProc , SAVE_PROGRESS);

 // Load in the status strings we will need if everything goes ok
 CResString WritingHeader (hInst , IDS_WRITINGHEADER);
 CResString WritingDir    (hInst , IDS_WRITINGDIRECTORY);
 CResString WritingFiles  (hInst , IDS_WRITINGFILES);

 SetWindowText (hwndTextCtrl , WritingHeader);

 if (!WriteNFIHeader (NewFile , Header)) {
     DeleteFile (TempFName);
     // Get rid of the progress dialog
     DestroyWindow (hwndProgressDlg);
     Filename[0] = '\0';
     return;
 }

 if (!WriteDirHeader (NewFile , DirHeader)) {
     DeleteFile (TempFName);
     // Get rid of the progress dialog
     DestroyWindow (hwndProgressDlg);
     Filename[0] = '\0';
     return;
 }

 // So far so good. Now, we want to create the directory. To do this, what
 // we do is loop through every single entry in the listview, since each one
 // represents a file. We then Output as much of the information in that
 // entry as we can. Later when we actually write the file section we go back
 // and update the parts that need it (such as the offset).
 WINASSERT (DirHeader.NumEntries == ListView_GetItemCount (hwndListview));

 SetWindowText (hwndTextCtrl , WritingDir);

 // These two vars are used in updating the process dialog. Since we could
 // possibly be writing directory entries a long time, we count them as
 // part of the operation. So, we make the total items twice as big as it
 // should be to start. Then writing all the entries will only update the
 // progress bar to around 50%.
 WPARAM Pct;
 int ThisItem = 0;
 int TotalItems = DirHeader.NumEntries;
 for (int Index = 0 ; Index < DirHeader.NumEntries ; Index++) {
     // Get the item with this index. The thing we're most interested in is
     // the lParam value and the image value. They have everything we need.
     LV_ITEM lvi;
     lvi.iItem = Index;
     lvi.iSubItem = 0;
     lvi.mask = LVIF_IMAGE | LVIF_PARAM;

     ListView_GetItem (hwndListview , &lvi);

     // We have a somewhat different job here depending on what image this
     // node is using. If the image is 0, then lParam is an actual Directory
     // entry ready to be written out. If the image is 1, then this is an
     // item which is to be added. In that case, we now have to create a
     // node based on the information that we have and write that. As per
     // usual on any error we delete the temporary file (which has already
     // been closed) and return.
     if (lvi.iImage == 0) {
         if (!WriteDirectoryEntry (NewFile , (NFIDirEntry *) lvi.lParam)) {
             DeleteFile (TempFName);
             // Get rid of the progress dialog
             DestroyWindow (hwndProgressDlg);
             Filename[0] = '\0';
             return;
         }
     } else {
         AddedFile *File = (AddedFile *) lvi.lParam;

         NFIDirEntry Entry;
         MakeRealEntry (File , Entry);

         // Now write the entry we have.
         if (!WriteDirectoryEntry (NewFile , &Entry)) {
             DeleteFile (TempFName);
             // Get rid of the progress dialog
             DestroyWindow (hwndProgressDlg);
             Filename[0] = '\0';
             return;
         }
     }

     if (ThisItem == TotalItems)
         Pct = 100;
     else
         Pct = (WPARAM) (((float) ThisItem / (float) TotalItems) * 100.0);

     ThisItem++;
     SendMessage (hwndProgressCtrl , PBM_SETPOS , Pct , 0);

     PumpMessages ();
 }

 // Flush everything written so far to disk.
 FlushFileBuffers (NewFile);

 SetWindowText (hwndTextCtrl , WritingFiles);

 ThisItem = 0;

 // Now, we loop through the items again. Only this time, we are actually
 // writing the files. We coudn't do this step before because all of the
 // directory entries have to be contiguous in the file.
 for (Index = 0 ; Index < DirHeader.NumEntries ; Index++) {
     // Get the item with this index. The thing we're most interested in is
     // the lParam value and the image value. They have everything we need.
     LV_ITEM lvi;
     lvi.iItem = Index;
     lvi.iSubItem = 0;
     lvi.mask = LVIF_IMAGE | LVIF_PARAM;

     ListView_GetItem (hwndListview , &lvi);

     // Now, depending on the image we have either a real node or an added
     // node. We use that to determine which helper function to call.
     if (lvi.iImage == 0) {
         if (!CopyNFIItem (NewFile , lvi)) {
             DeleteFile (TempFName);
             // Get rid of the progress dialog
             DestroyWindow (hwndProgressDlg);
             Filename[0] = '\0';
             return;
         }
     } else {
         if (!AddFileToNFI (NewFile , lvi)) {
             DeleteFile (TempFName);
             // Get rid of the progress dialog
             DestroyWindow (hwndProgressDlg);
             Filename[0] = '\0';
             return;
         }
     }

     if (ThisItem == TotalItems)
         Pct = 100;
     else
         Pct = (WPARAM) (((float) ThisItem / (float) TotalItems) * 100.0);

     ThisItem++;
     SendMessage (hwndProgressCtrl , PBM_SETPOS , Pct , 0);

     PumpMessages ();
 }

 // If we get this far then by jove we must have succeded! The first thing we
 // want to do is close the real main NFI File.  We then try to delete whatever
 // file we where supposed to save the NFI under. If we where doing a Save As
 // then nothing important will happen. If we are doing a save, then this will
 // nuke the original file, since we now have a new one in it's place. We also
 // close the temporary file.
 CloseHandle (MainNFIFile);
 CloseHandle (NewFile);
 DeleteFile (Filename);

 // Now, rename the temporary file to the real file.
 MoveFile (TempFName , Filename);

 // Get rid of the progress dialog
 DestroyWindow (hwndProgressDlg);

 // Now, we Reopen the file.
 DoOpen (FALSE);
}

/*****************************************************************************
 * Name         - DoSaveAs
 * Purpose      - This function prompts the user for a filename, and then
 *                calls the DoSave function in order to actually save the
 *                file.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DoSaveAs (void)
{
 OPENFILENAME of;
 char         FileToSave[MAX_PATH]="";
 CResString   Title  (hInst , IDS_SAVEFILETITLE);
 CResString   Filter (hInst , IDS_FILTERSTRING);
 CResString   DefExt (hInst , IDS_DEFAULTEXTENSION);

 // Since you cannot use the \0 notation in a resource string (it stops loading
 // there), we specify the filter using | marks intead of null characters.
 // Here we get a temporary pointer to the string data and scan through it
 // converting all of the | characters to null ones.
 char *Temp = Filter;
 while (*Temp) {
     if (*Temp == '|')
         *Temp = '\0';
     Temp++;
 }

 // Set up the control struct
 of.lStructSize       = sizeof (of);
 of.hwndOwner         = hwndMain;
 of.hInstance         = hInst;
 of.lpstrFilter       = Filter;
 of.lpstrCustomFilter = NULL;
 of.nFilterIndex      = 0;
 of.lpstrFile         = FileToSave;
 of.nMaxFile          = sizeof (FileToSave);
 of.lpstrFileTitle    = NULL;
 of.lpstrInitialDir   = NULL;
 of.lpstrTitle        = Title;
 of.lpstrDefExt       = DefExt;
 of.Flags             = OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY |
                        OFN_PATHMUSTEXIST;

 // Prompt the user to pick a filename. If the user doesn't pick one, then
 // just return.
 if (!GetSaveFileName (&of))
     return;

 // Set the filename to the one the user selected and then call the save
 // function
 strcpy (Filename , FileToSave);
 DoSave ();
}

/*****************************************************************************
 * Name         - ReadErrorAndReset
 * Purpose      - This is a helper function called by the DoOpen function.
 *                Whenever there is an error reading or opening the file, this
 *                is called to display a message box, and then put the program
 *                into a state where there is a new, untitled document loaded.
 * Arguments    - ErrorId: Res ID of the string used for the error message
 * Return Value - None
 ****************************************************************************/
void ReadErrorAndReset (int ErrorId)
{
 // Display the error message
 CResString Title (hInst , IDS_ERRORREADTITLE);
 CResString Msg   (hInst , ErrorId);
 MessageBox (hwndMain , Msg , Title , MB_OK | MB_ICONERROR);

 // Reset the file and filename
 DoNew ();
 Filename[0] = '\0';
 MergeFilename (hwndMain , szTitle , Filename);
}

/*****************************************************************************
 * Name         - GetDirectoryHeader
 * Purpose      - This loads the directory header from the main NFI file.
 * Arguments    - Header: Struct to store the header info in.
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL GetDirectoryHeader (NFIDirHeader *Header)
{
 DWORD BytesRead;

 if (!ReadFile (MainNFIFile , &(Header->NumEntries) ,
                sizeof (Header->NumEntries) , &BytesRead , NULL) ||
                BytesRead != sizeof (Header->NumEntries)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }
 if (!ReadFile (MainNFIFile , &(Header->Reserved) ,
                sizeof (Header->Reserved) , &BytesRead , NULL) ||
                BytesRead != sizeof (Header->Reserved)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }

 return TRUE;
}

/*****************************************************************************
 * Name         - GetDirectoryEntry
 * Purpose      - This loads the next directory entry from the main NFI file.
 * Arguments    - Entry: Struct to store the entry info in.
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL GetDirectoryEntry (NFIDirEntry *Entry)
{
 DWORD BytesRead;

 if (!ReadFile (MainNFIFile , Entry->Name ,
                sizeof (Entry->Name) , &BytesRead , NULL) ||
                BytesRead != sizeof (Entry->Name)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }
 if (!ReadFile (MainNFIFile , &(Entry->Offset) ,
                sizeof (Entry->Offset) , &BytesRead , NULL) ||
                BytesRead != sizeof (Entry->Offset)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }
 if (!ReadFile (MainNFIFile , &(Entry->Length) ,
                sizeof (Entry->Length) , &BytesRead , NULL) ||
                BytesRead != sizeof (Entry->Length)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }
 if (!ReadFile (MainNFIFile , &(Entry->Reserved[0]) ,
                sizeof (Entry->Reserved[0]) , &BytesRead , NULL) ||
                BytesRead != sizeof (Entry->Reserved[0])) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }
 if (!ReadFile (MainNFIFile , &(Entry->Reserved[1]) ,
                sizeof (Entry->Reserved[1]) , &BytesRead , NULL) ||
                BytesRead != sizeof (Entry->Reserved[1])) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return FALSE;
 }

 return TRUE;
}

/*****************************************************************************
 * Name         - AddNFIFileItem
 * Purpose      - Adds Entry to the linked list of directory entries and then
 *                inserts a listview item representing it.
 * Arguments    - Entry: Entry used for this directory entry.
 *                Index: Into to insert this item at
 * Return Value - None
 ****************************************************************************/
void AddNFIFileItem (NFIDirEntry *Entry , int Index)
{
 // First, we add ENtry to the linked list. This is easy. If we don't have
 // a head, this is it. Otherwise, this entry's Next is the head, then we make
 // it the head.
 if (DirectoryRoot == NULL)
     DirectoryRoot = Entry;
 else {
     // The next of this node points to what the head is now
     Entry->Next = DirectoryRoot;
     // The prev of the head node points at us
     DirectoryRoot->Prev = Entry;
     // The head of thelist is now us
     DirectoryRoot = Entry;
 }

 static LV_ITEM lvi;
 static char Buffer[_MAX_PATH + 10];
 static char Path[_MAX_DIR];
 static char FName[_MAX_FNAME];
 static char Ext[_MAX_EXT];

 // Split the filename we got into it's parts, then make the filename portion
 // one string.
 _splitpath  (Entry->Name , NULL , Path , FName , Ext);
 strcpy (Buffer , FName);
 strcat (Buffer , Ext);

 lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;

 lvi.iItem    = Index;
 lvi.iSubItem = 0;
 lvi.pszText  = Buffer;
 lvi.iImage   = 0;
 lvi.lParam   = (LPARAM) Entry;

 // Insert the item
 ListView_InsertItem  (hwndListview , &lvi);

 // Set the size text
 ultoa (Entry->Length , Buffer , 10);
 ListView_SetItemText (hwndListview , Index , 1 , Buffer);

 // Set the offset text
 ultoa (Entry->Offset , Buffer , 10);
 ListView_SetItemText (hwndListview , Index , 2 , Buffer);

 // Add the path.
 ListView_SetItemText (hwndListview , Index , 3 , Path);
}

/*****************************************************************************
 * Name         - AddPathToTree
 * Purpose      - The program keeps a list of the unique paths inside the
 *                NFI file when it is loaded and updated and uses that list
 *                to be able to populate some dialog box controls. A binary
 *                tree is used to store the information since it is accessed
 *                often when it is constructed.
 * Arguments    - Under: The node under which to add the new node
 *                Path: The new path to add
 * Return Value - Returns a pointer to the node that Path was added under.
 *                This always ends up being the root of the tree when called
 *                elsewhere, but the function calls itself recursively and
 *                so it is needed
 ****************************************************************************/
PathInfo *AddPathToTree (PathInfo *Under , char *Path)
{
 WINASSERT (Path);
 int Condition;

 // Is the node we're supposed to add under null? If it is, then we must make
 // a new node and init it
 // then init it's fields
 if (Under == NULL) {
     Under = (PathInfo *) new char[sizeof (PathInfo)];
     memset (Under , 0 , sizeof (PathInfo));

     Under->Path = new char[strlen (Path) + 1];
     strcpy (Under->Path , Path);
 } else {
     WINASSERT (Under->Path);

     // Compare the path inormation. Note this is case insensitive since files
     // in the NFI are not case sensitive
     Condition = strcmpi (Path , Under->Path);

     // Depending on if the new path is less than or greater than this one,
     // make a recursive call down the proper branch of the tree. If the two
     // are equal nothing happens since we don't care about duplicate entries.
     if (Condition < 0)
         Under->Left = AddPathToTree (Under->Left , Path);
     else
         if (Condition > 0)
             Under->Right = AddPathToTree (Under->Right , Path);
 }

 return Under;
}

/*****************************************************************************
 * Name         - DeleteNode
 * Purpose      - Releases the memory of a path info node. It first gets rid
 *                of the path information, then the node. Notice that we
 *                case the node to a char pointer and delete that since that
 *                is how it was allocated.
 * Arguments    - Node: Node to delete
 * Return Value - None
 ****************************************************************************/
void DeleteNode (PathInfo *Node)
{
 delete[] Node->Path;
 char *Temp = (char *) Node;
 delete[] Temp;
}

/*****************************************************************************
 * Name         - DeleteChildNodes
 * Purpose      - Gives a node in the path information tree, this function
 *                will delete all of the child nodes under it. It does this
 *                by first recursing as far down the tree as it can. Once
 *                there, it looks at the node to it's left, to see if it has
 *                any children. If it doesn't, then the left node is deleted.
 *                a similar operatiopn occurs on the right. Since this is
 *                recursive, if there is a child node, it's children will
 *                get deleteed on the way back up the stack. The function
 *                does not actually delete the passed in node, only it's
 *                children. If a child is deleted, the pointer to that child
 *                is set to NULL.
 * Arguments    - Node: Node to work on
 * Return Value - None
 ****************************************************************************/
void DeleteChildNodes (PathInfo *Node)
{
 PathInfo *ThisLeft;
 PathInfo *ThisRight;

 // Only do something if we actually have a node.
 if (Node) {
     // Recurse as far down the tree as we can at the moment.
     DeleteChildNodes (Node->Left);
     DeleteChildNodes (Node->Right);

     // Get pointers to the nodes that are left and right of this one.
     ThisLeft = Node->Left;
     ThisRight = Node->Right;

     // Now, if the left node has no children, we can delete it.
     if (ThisLeft && ThisLeft->Left == NULL && ThisLeft->Right == NULL) {
         // Before we do anything, mark this node as having no left child.
         // Then delete it
         Node->Left = NULL;
         DeleteNode (ThisLeft);
     }

     // Similarly, if the right node has no children, delete it
     if (ThisRight && ThisRight->Left == NULL && ThisRight->Right == NULL) {
         // Before we do anything, mark this node as having no right child.
         Node->Right = NULL;
         DeleteNode (ThisRight);
     }
 }
}

/*****************************************************************************
 * Name         - DeletePathTree
 * Purpose      - This function deletes the path tree. It first deletes all
 *                of the child nodes of the root of the tree. It then deletes
 *                the tree itself.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DeletePathTree (void)
{
 DeleteChildNodes (NFIPathInfo);
 DeleteNode (NFIPathInfo);
 NFIPathInfo = NULL;
}

/*****************************************************************************
 * Name         - DoActualOpenOperation
 * Purpose      - This is the function that actually does the loading of the
 *                NFI file. It assumes that the global MainNFIFile handle
 *                is a valid value and uses it to open the file. It also
 *                assumes that the global handle to the progress dialog box
 *                has been set.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DoActualOpenOperation (void)
{
 WINASSERT (MainNFIFile != INVALID_HANDLE_VALUE);
 WINASSERT (hwndProgressDlg);

 // Ok the file is open, try to read in the header of the file. Just in case
 // the byte alignment is not 1, we read in one structure member at a time.
 DWORD     BytesRead;
 NFIHeader Header;

 WINASSERT (hwndProgressCtrl);
 WINASSERT (hwndTextCtrl);

 // Load in all of the status resource strings that we'll need.
 CResString ReadingHeader (hInst , IDS_READINGHEADER);
 CResString ReadingDir    (hInst , IDS_READINGDIRECTORY);

 // We start by reading the header, so tell the static control to display that
 // message.
 SetWindowText (hwndTextCtrl , ReadingHeader);

 // Get the header tag
 if (!ReadFile (MainNFIFile , Header.IDTag ,
                sizeof (Header.IDTag) , &BytesRead , NULL) ||
                BytesRead != sizeof (Header.IDTag)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return;
 }

 // Is this an NFI File?
 if (strcmp (Header.IDTag , NFI_HEADERID) != 0) {
     ReadErrorAndReset (IDS_ERRORNOTNFI);
     return;
 }

 // Get the version information
 if (!ReadFile (MainNFIFile , &(Header.Version) ,
                sizeof (Header.Version) , &BytesRead , NULL) ||
                BytesRead != sizeof (Header.Version)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return;
 }

 // Is the version OK?
 if (Header.Version != NFI_VERSION) {
     ReadErrorAndReset (IDS_ERRORBADVER);
     return;
 }

 // Get the rest of the header
 if (!ReadFile (MainNFIFile , &(Header.Reserved) ,
                sizeof (Header.Reserved) , &BytesRead , NULL) ||
                BytesRead != sizeof (Header.Reserved)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return;
 }
 if (!ReadFile (MainNFIFile , &(Header.DirOffset) ,
                sizeof (Header.DirOffset) , &BytesRead , NULL) ||
                BytesRead != sizeof (Header.DirOffset)) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return;
 }

 // Now that we have the offset of the directory, go there
 if (SetFilePointer (MainNFIFile , Header.DirOffset , NULL , FILE_BEGIN) == -1) {
     ReadErrorAndReset (IDS_ERRORREADING);
     return;
 }

 // Get the directory header
 NFIDirHeader DirHeader;
 if (!GetDirectoryHeader (&DirHeader))
     return;

 // We now know how many files there are in this NFI file. First thing, set
 // the global varible.
 FilesInNFI = DirHeader.NumEntries;

 // The header has been completely loaded now. We're loading the directory.
 // First, update the status text.
 SetWindowText (hwndTextCtrl , ReadingDir);

 // Tell the ListView control how many items we are going to be inserting.
 // This allows it to resize it's internal data structures just once instead
 // of a bunch of times. Much more efficient.
 ListView_SetItemCount (hwndListview , FilesInNFI);

 // Tell the ListView control not to redraw itself as we fool with it. Also
 // for speed reasons.
 SendMessage (hwndListview , WM_SETREDRAW , (WPARAM) FALSE , 0);

 // Now, we know how many files there are. So we loop, reading that many
 // directory entries. For each entry we make a dynamic structure, initalize
 // it with the data, and pass it off to a helper function for insertion
 // into the listview control.

 WPARAM Pct;
 for (long Index = 0 ; Index < FilesInNFI ; Index++) {
     // Allocate an entry and zero init it
     NFIDirEntry *Entry = (NFIDirEntry *) new char[sizeof (NFIDirEntry)];
     memset (Entry , 0 , sizeof (NFIDirEntry));

     // Read it in
     if (!GetDirectoryEntry (Entry)) {
         SendMessage (hwndListview , WM_SETREDRAW , (WPARAM) TRUE , 0);
         InvalidateRect (hwndListview , NULL , FALSE);
         UpdateWindow (hwndListview);
         return;
     }

     // Now, add it to the internal lists, as well as the listview control.
     AddNFIFileItem (Entry , Index);

     // Split the filename apart to get the directory information off of it.
     // We then try to add that to the binary tree that represents all of the
     // paths in the tree.
     static char Buffer[_MAX_PATH + 10];
     static char Path[_MAX_DIR];

     // Get the path part of the filename. Then, prepend a leading / to it.
     _splitpath  (Entry->Name , NULL , Path , NULL , NULL);
     strcpy (Buffer , Path);

     NFIPathInfo = AddPathToTree (NFIPathInfo , Buffer);

     // Now we need to update the progress control. We need to find the % that
     // we are done. This is the file we're at divided by the total and then
     // multiplied by 100. We case the numbers to float during the operation
     // and then back to a WPARAM afterwards. We then tell the progress control
     // to use that number as it's position. Note that If we determine this is
     // the last file to be loaded, we explicitly set the percent to 100. This
     // takes care of rounding errors.
     if (Index == FilesInNFI - 1)
         Pct = 100;
     else
         Pct = (WPARAM) (((float) Index / (float) FilesInNFI) * 100.0);

     SendMessage (hwndProgressCtrl , PBM_SETPOS , Pct , 0);

     PumpMessages ();
 }

 // Now turn redraw back on and get it to repaint
 SendMessage (hwndListview , WM_SETREDRAW , (WPARAM) TRUE , 0);
 InvalidateRect (hwndListview , NULL , FALSE);
 UpdateWindow (hwndListview);

 // We leave the file open until the user says they want a new one or quit.
 // This makes sure that nothing happens to the file while the user it working
 // on it (like say deleting it) which would cause problems when adding new
 // files to it.

 // Update the statusbar with the new information. Notice that here we do
 // not bother to update the selected count, since idle processing handles
 // that. We just make sure that the file is marked as not modified and that
 // we update the count of files that are in the NFI file.
 SetFileStatus (FALSE);
 FilesToAdd = 0;
 UpdateFileCount ();

 // Add this file to the shell's list of recently open documents so that it
 // can be accessed again quickly.
 SHAddToRecentDocs (SHARD_PATH , Filename);
}

/*****************************************************************************
 * Name         - DoOpen
 * Purpose      - This functions performs an open command. It assumes that
 *                the Filename member is valid, and tries to open and read
 *                in a new file. It assumes that all of the data structures
 *                that represent a file are empty (i.e. non existant). What it
 *                does, is it first gets the file open and ready to go. Once
 *                it has done that, it engages the Progress Dialog box with
 *                a call to DialogBoxParam. That function will not return until
 *                the dialog box closes itself. After the dialog box
 *                initializes itself though, it will process any messages sent
 *                to it and then go idle. When it does that, it notifies it's
 *                owner (the main window frame). We take that opportunity to
 *                call the function that performs all of the file work. That
 *                is DoActualOpenOperation. Once THAT has completed, it will
 *                return to the main window handler for the idle message. That
 *                function will then pass a message to the dialog box to tell
 *                it to close itself. The dialog box will call EndDialog in
 *                order to shut itself down. When that happens, control will
 *                be transferred back to here, to the line after the one that
 *                started the dialog box. We then just return.
 * Arguments    - PromptForOpen: TRUE if the function should prompt the user
 *                               for the file to load, FALSE if it should just
 *                               take it that Filename is a semi-valid filename.
 * Return Value - None
 ****************************************************************************/
void DoOpen (BOOL PromptForOpen)
{
 // Prompt the user for the file to open. If they don't specify one, we just
 // return without doing anything.
 char FileToOpen[MAX_PATH]="";

 if (PromptForOpen) {
     OPENFILENAME of;
     CResString   Title  (hInst , IDS_OPENFILETITLE);
     CResString   Filter (hInst , IDS_FILTERSTRING);
     CResString   DefExt (hInst , IDS_DEFAULTEXTENSION);

     // Since you cannot use the \0 notation in a resource string (it stops loading
     // there), we specify the filter using | marks intead of null characters.
     // Here we get a temporary pointer to the string data and scan through it
     // converting all of the | characters to null ones.
     char *Temp = Filter;
     while (*Temp) {
         if (*Temp == '|')
             *Temp = '\0';
         Temp++;
     }

     of.lStructSize       = sizeof (of);
     of.hwndOwner         = hwndMain;
     of.hInstance         = hInst;
     of.lpstrFilter       = Filter;
     of.lpstrCustomFilter = NULL;
     of.nFilterIndex      = 0;
     of.lpstrFile         = FileToOpen;
     of.nMaxFile          = sizeof (FileToOpen);
     of.lpstrFileTitle    = NULL;
     of.lpstrInitialDir   = NULL;
     of.lpstrTitle        = Title;
     of.lpstrDefExt       = DefExt;
     of.Flags             = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY |
                            OFN_PATHMUSTEXIST;
     if (!GetOpenFileName (&of))
         return;
 } else
     strcpy (FileToOpen , Filename);

 // Now that we know we are opening up a file, put the program into a state
 // where it is ready to open a file.
 DoNew ();

 // Now we are ready to actually load the file. If we should encounter a
 // problem at any time, we will alert the user using a message box, and then
 // put the program into a state as if the user had clicked the "new" button.
 // This stops us from having a partially loaded file in memory.
 WINASSERT (MainNFIFile == INVALID_HANDLE_VALUE);

 // First up, try to open the file. We open the file for read access only.
 MainNFIFile = CreateFile (FileToOpen , GENERIC_READ , FILE_SHARE_READ ,
                           NULL , OPEN_EXISTING , FILE_FLAG_RANDOM_ACCESS ,
                           NULL);

 // Was there an error? If there was, we display an error message and bail.
 if (MainNFIFile == INVALID_HANDLE_VALUE) {
     ReadErrorAndReset (IDS_ERROROPENING);
     return;
 }

 // Update the filename now.
 strcpy (Filename , FileToOpen);
 MergeFilename (hwndMain , szTitle , Filename);

 // The file was opened, so create the dialog box and get things rolling.
 WINASSERT (hwndProgressDlg == NULL);
 CreateDialogParam (hInst, MAKEINTRESOURCE (IDD_PROGRESSDLG) , hwndMain , (DLGPROC) ProgressProc , OPEN_PROGRESS);

 // Do the actual operation
 DoActualOpenOperation ();

 // Get rid of the progress dialog
 DestroyWindow (hwndProgressDlg);
}

/*****************************************************************************
 * Name         - DoNew
 * Purpose      - This function performs a new command by clearing out all of
 *                the information the program uses to represent an open file
 *                and resetting it to it's startup state. This is done
 *                unconditionally
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DoNew (void)
{
 // If the main file is open, close it
 if (MainNFIFile != INVALID_HANDLE_VALUE) {
     CloseHandle (MainNFIFile);
     MainNFIFile = INVALID_HANDLE_VALUE;
 }

 // Make the progress dialog box
 WINASSERT (hwndProgressDlg == NULL);
 CreateDialogParam (hInst, MAKEINTRESOURCE (IDD_PROGRESSDLG) , hwndMain , (DLGPROC) ProgressProc , NEW_PROGRESS);

 CResString Status (hInst , IDS_CLEARINGMEM);
 SetWindowText (hwndTextCtrl , Status);

 // Turn off redraw for the control so we can fiddle with the icons without
 // it going nuts.
 SendMessage (hwndListview , WM_SETREDRAW , (WPARAM) FALSE , 0);

 int Index = 0;
 int Total = ListView_GetItemCount (hwndListview);
 WPARAM Pct;
 while (Index < Total) {
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);
     ListView_DeleteItem (hwndListview , 0);

     if (Index == Total)
         Pct = 100;
     else
         Pct = (WPARAM) (((float) Index / (float) Total) * 100.0);

     SendMessage (hwndProgressCtrl , PBM_SETPOS , Pct , 0);

     PumpMessages ();

     Index += 10;
 }

 // Now turn redraw back on and get it to repaint
 SendMessage (hwndListview , WM_SETREDRAW , (WPARAM) TRUE , 0);
 InvalidateRect (hwndListview , NULL , FALSE);
 UpdateWindow (hwndListview);

 // Set the file unmodified and reset variabes
 SetFileStatus (FALSE);
 FilesInNFI = 0;
 FilesToAdd = 0;
 UpdateFileCount ();

 // Nukify the linked list of loaded directory items.
 NFIDirEntry *Entry = DirectoryRoot;
 while (Entry) {
     // Get a char pointer to this node (we allocated the memory as an
     // array of chars.
     char *Temp = (char *) Entry;

     // Advance to the next entry in the list before we delete this node.
     Entry = Entry->Next;

     // Nuke the memory for this entry
     delete[] Temp;
 }
 DirectoryRoot = NULL;

 // Nukify the linked list of files to be added.
 DestroyAddedFileList ();

 DestroyWindow (hwndProgressDlg);
}

/*****************************************************************************
 * Name         - CreateFilePath
 * Purpose      - Given a fully qualified filename (this includes a drive and
 *                a path), the function attempts to create the directory
 *                structure which is needed to save the file in that
 *                directory.
 * Arguments    - Filename: Fully qualified pathname to use
 * Return Value - None
 ****************************************************************************/
void CreateFilePath (char *Filename)
{
 char Dir[_MAX_DRIVE + MAX_PATH];
 char Path[MAX_PATH];
 char PathBefore[MAX_PATH]="";

 // Split apart the filename from the directory and path
 _splitpath (Filename , Dir , Path , NULL , NULL);
 strcat (Dir , Path);
 strcpy (Path , Dir);

 // If the last character of the path isn't a \,then add one, just in case.
 if (Path[strlen (Path) - 1] != '\\')
     strcat (Path , "\\");

 // If the Path is a single \ then the path represents the root and does
 // not need to be created (and can't either) so just quit.
 if (strcmp (Path , "\\") == 0)
     return;

 // If the length of the path is 3, the second character is a colon and the
 // last is a \, then also return that means root too.
 if (strlen (Path) == 3 && Path[1] == ':' && Path[2] == '\\')
     return;

 // Once we get here, we know that Path is not just "\" or "J:\" . It must be
 // at least "\something\" or "j:\something\". What we do here is pick off
 // one subdirectory at a time and create it. The variable PathBefore is the
 // path we know is valid up to this point. We we get a subdir we make add it
 // to the end of PathBefore and then create that subdir. We continue until
 // the whole string as been processed. To start we copy everything from the
 // beginning of the string, up to and including the first \ in the string
 // into PathBefore.
 char *FirstSlash = strchr (Path , '\\');

 WINASSERT (FirstSlash);

 // FirstSlash now points to the first \ in Path. Make it a NULL charcter and
 // then advance it.
 *FirstSlash = '\0';
 FirstSlash++;

 // Now copy Path into PathBefore and append a \.
 strcpy (PathBefore , Path);
 strcat (PathBefore , "\\");

 // Keep looping forever until the loop decides to break.
 char *NextSlash;
 do {
     // FirstSlash points to the first part of the subdirectory that we want
     // to process. Find out where it ends.
     NextSlash = strchr (FirstSlash , '\\');
     WINASSERT (NextSlash);

     // Once again make it a NULL then advance it.
     *NextSlash = '\0';
     NextSlash++;

     // Now FirstSlash points to the first character in the subdirectory and
     // the place where the next slash is has been altered so that it is a
     // null. That make FirstSlash point to a string which is the next
     // subdir. Tack that onto PathBefore.
     strcat (PathBefore , FirstSlash);

     // Now PathBefore is the name of a subdirectory to create. Do it
     CreateDirectory (PathBefore , NULL);

     // Now make FirstSlash equal to NextSLash for the next pass through the
     // loop, and add a \ to the end of PathBefore for next time as well.
     FirstSlash = NextSlash;
     strcat (PathBefore , "\\");
 } while (*FirstSlash);
}

/*****************************************************************************
 * Name         - TweakExtractedFilename
 * Purpose      - This function tweaks the given filename in accordance with
 *                how the user wants the filename to be extracted to the disk.
 * Arguments    - Name: Filename to tweak
 * Return Value - None
 ****************************************************************************/
void TweakExtractedFilename (char *Name)
{
 WINASSERT (Name);

 // Now, are spaces in filenames supposed to be converted to underscores?
 if (Setup.ExtractSpaceProcess == 1) {
     // Scan through the string, converting all spaces to underscores.
     char *Temp = Name;
     while (*Temp) {
         if (*Temp == ' ')
             *Temp = '_';
         Temp++;
     }
 } else
     // No. Should spaces be deleted?
     if (Setup.ExtractSpaceProcess == 2) {
         // We need to scan through the string taking out any spaces. To
         // do that we first make a temporary buffer and copy the string into
         // it.
         char TempBuffer[MAX_PATH];
         strcpy (TempBuffer , Name);

         // make a source index and a destination index and set both to 0.
         int SrcIndex=0;
         int DstIndex=0;

         // Now, scan through the source (The temp buffer), incrementing
         // SrcIndex each time, until we get to the last character of the
         // string. For each character, if it is not a space, we put it into
         // File->InternalName[DstIndex] and update the destination index.
         while (1) {
             if (TempBuffer[SrcIndex] != ' ')
                 Name[DstIndex++] = TempBuffer[SrcIndex];

             if (TempBuffer[SrcIndex] == '\0')
                 break;
             else
                 SrcIndex++;
         }
     }

 // Scan through the string now and make sure that all of the / are converted
 // to a \.
 char *Temp = Name;
 while (*Temp) {
     if (*Temp == '/')
         *Temp = '\\';
     Temp++;
 }
}

/*****************************************************************************
 * Name         - ExtractionFileOperation
 * Purpose      - This function performs the meat of an extraction operation.
 *                It copys Length bytes from Source to Dest. If there is any
 *                Error, the handle referred to by Dest is closed before the
 *                function returns.
 * Arguments    - Source: File to copy from
 *                Dest: File to copy to
 *                Length: Bytes to copy
 *                FName: Filename to use in error messages.
 * Return Value - TRUE on success, FALSE on failure.
 ****************************************************************************/
BOOL ExtractionFileOperation (HANDLE Source , HANDLE Dest , LONG Length ,
                              char *FName)
{
 // Now, copy all of the data from that file to the new file. First, determine
 // how many bytes we have to read.
 LONG BytesToRead = Length;
 LONG BytesThisLoop;

 // Loop until all the bytes have been read and copied over.
 while (BytesToRead) {
     // Determine how many bytes we should go this chunk.
     if (BytesToRead > MAX_FILE_BUFFER)
         BytesThisLoop = MAX_FILE_BUFFER;
     else
         BytesThisLoop = BytesToRead;

     // Copy the data. Return on error, the new file will already have been
     // closed.
     if (!CopyDataChunk (Source , Dest , BytesThisLoop , IDS_ERROREXTRACTTITLE , IDS_ERROREXTRACTING , FName))
         return FALSE;

     // Update the bytes left to read.
     BytesToRead -= BytesThisLoop;
 }

 return TRUE;
}

/*****************************************************************************
 * Name         - ExtractEntry
 * Purpose      - This function performs the actual job of extracting a file
 *                from the NFI File. It takes as an input on the file to
 *                extract a listview item which specifies all the ifnormation
 *                needed to extract the file. If the entry specifies an
 *                entry which is in the NFI file, it is extracted. If it
 *                specifies an entry which is waiting to be added, then the
 *                file is copied from it's present location to the extraction
 *                location.
 * Arguments    - Item: Listview item containing the information of the file
 *                      to be extracted.
 *                ToTemp: if TRUE the file is extracted to the temporary
 *                        directory. If FALSE, the file is extracted to the
 *                        user specified extraction directory.
 *                ExtractedName: Pointer to a buffer that is filled with the
 *                               name of the extracted file. If NULL, the
 *                               information is not recorded.
 * Return Value - TRUE on success, FALSE on failure.
 ****************************************************************************/
BOOL ExtractEntry (LV_ITEM &Item , BOOL ToTemp , char *ExtractedName)
{
 char OutName[MAX_PATH]="";

 // If image is 0 this is a loaded entry, if 1 it is an added entry.
 if (Item.iImage == 0) {
     // This is a loaded entry.
     NFIDirEntry *Entry = (NFIDirEntry *) Item.lParam;

     // Split the name for this entry into a path and filename portion.
     char PathPart[MAX_PATH];
     char FilePart[_MAX_FNAME + _MAX_EXT];
     char ExtPart [_MAX_EXT];

     _splitpath (Entry->Name , NULL , PathPart , FilePart , ExtPart);
     strcat (FilePart , ExtPart);

     // Determine the complete output filename that should be used to create
     // the extracted file. If ToTemp is true then we save the file in the
     // temporary directory. Otherwise, we save it to the extraction directory,
     // possibly maintaining directory structure.
     if (ToTemp) {
        // Get the temporary path and tack on the filename.
        GetTemporaryDirectory (OutName);
        strcat (OutName , FilePart);
     } else {
         // Get the extraction directory.
         strcpy (OutName , Setup.ExtractDir);

         // Temporarily store the internal path and filename so we can perform
         // processing on it before we add the drive and directory that is
         // actually being used.
         char InternalFileAndPath[MAX_PATH]="";

         // If we are supposed to maintain directory structure, tack on the
         // file's path. Notice that we add +1 to the path pointer so that the
         // initial / doesn't get added on. THe extraction directory already
         // has a trailing / on the end.
         if (Setup.MaintainDirStructure) {
             if (strlen (PathPart) > 1)
                 strcpy (InternalFileAndPath , PathPart + 1);
         }

         // Add the filename.
         strcat (InternalFileAndPath , FilePart);
         TweakExtractedFilename (InternalFileAndPath);
         strcat (OutName , InternalFileAndPath);
     }

     CreateFilePath (OutName);

     if (ExtractedName)
         strcpy (ExtractedName , OutName);

     // Seek to the proper place in the main file where this file starts.
     SetFilePointer (MainNFIFile , Entry->Offset , NULL , FILE_BEGIN);

     // Try to create the file what we will be creating.
     HANDLE ExtractedFile = CreateFile (OutName , GENERIC_WRITE , 0 , NULL ,
                                        CREATE_ALWAYS ,
                                        FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS ,
                                        NULL);

     if (ExtractedFile == INVALID_HANDLE_VALUE) {
         CResString Title   (hInst , IDS_ERROREXTRACTTITLE);
         CResString Message (hInst , IDS_ERRORCREATINGEXT);
         char Buffer[MAX_PATH];
         wsprintf (Buffer , Message , OutName);

         MessageBox (hwndMain , Buffer , Title , MB_OK | MB_ICONERROR);
         return FALSE;
     }

     // Copy the file over.
     if (!ExtractionFileOperation (MainNFIFile , ExtractedFile ,
                                   Entry->Length , OutName)) {
         DeleteFile (OutName);
         return FALSE;
     }

     // Close the handle to the current file since we where successful.
     CloseHandle (ExtractedFile);
 } else {
     // This is an added item.
     AddedFile *File = (AddedFile *) Item.lParam;

     // The output name is constructed the same way for this as for a loaded
     // directory entry, with the advantage that we don't need to split the
     // path to get the info we want.
     if (ToTemp) {
        // Get the temporary path and tack on the filename.
        GetTemporaryDirectory (OutName);
        strcat (OutName , File->InternalName);
     } else {
         // Get the extraction directory.
         strcpy (OutName , Setup.ExtractDir);

         // Temporarily store the internal path and filename so we can perform
         // processing on it before we add the drive and directory that is
         // actually being used.
         char InternalFileAndPath[MAX_PATH]="";

         // If we are supposed to maintain directory structure, tack on the
         // file's path. Notice that we add +1 to the path pointer so that the
         // initial / doesn't get added on. THe extraction directory already
         // has a trailing / on the end.
         if (Setup.MaintainDirStructure) {
             if (strlen (File->InternalPath) > 1)
                 strcat (InternalFileAndPath , File->InternalPath + 1);
         }

         // Add the filename.
         strcat (InternalFileAndPath , File->InternalName);
         TweakExtractedFilename (InternalFileAndPath);
         strcat (OutName , InternalFileAndPath);
     }

     CreateFilePath (OutName);

     if (ExtractedName)
         strcpy (ExtractedName , OutName);

     // Try to create the file what we will be creating.
     HANDLE ExtractedFile = CreateFile (OutName , GENERIC_WRITE , 0 , NULL ,
                                        CREATE_ALWAYS ,
                                        FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS ,
                                        NULL);

     if (ExtractedFile == INVALID_HANDLE_VALUE) {
         CResString Title   (hInst , IDS_ERROREXTRACTTITLE);
         CResString Message (hInst , IDS_ERRORCREATINGEXT);
         char Buffer[MAX_PATH];
         wsprintf (Buffer , Message , OutName);

         MessageBox (hwndMain , Buffer , Title , MB_OK | MB_ICONERROR);
         return FALSE;
     }

     // Try to Open the file we will be copying from
     HANDLE SourceFile = CreateFile (File->AddFrom , GENERIC_READ ,
                                     FILE_SHARE_READ , NULL , OPEN_EXISTING ,
                                     FILE_FLAG_RANDOM_ACCESS , NULL);

     if (SourceFile == INVALID_HANDLE_VALUE) {
         CResString Title   (hInst , IDS_ERROREXTRACTTITLE);
         CResString Message (hInst , IDS_ERROROPENINGEXT);
         char Buffer[MAX_PATH];
         wsprintf (Buffer , Message , File->AddFrom);

         CloseHandle (ExtractedFile);
         DeleteFile (OutName);

         MessageBox (hwndMain , Buffer , Title , MB_OK | MB_ICONERROR);
         return FALSE;
     }

     // Copy the file over.
     if (!ExtractionFileOperation (SourceFile , ExtractedFile ,
                                   GetFileSize (SourceFile , NULL) , OutName)) {
         CloseHandle (SourceFile);
         DeleteFile (OutName);
         return FALSE;
     }

     // Close both handles we must have succeeded.
     CloseHandle (SourceFile);
     CloseHandle (ExtractedFile);
 }

 // If we get here we succeeded.
 return TRUE;
}

/*****************************************************************************
 * Name         - DoExtraction
 * Purpose      - This function performs the actual extraction operation. It
 *                gets the data it needs to do it's job from the listview
 *                control and the global setup information.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void DoExtraction (void)
{
 int TotalItems;

 int AllItems = ListView_GetItemCount (hwndListview);
 int SelItems = ListView_GetSelectedCount (hwndListview);
 // Find out the Total number of files that are going to be extracted.
 switch (ExtractProcess) {
     case 0: TotalItems = AllItems; break;
     case 1: TotalItems = SelItems; break;
     case 2: TotalItems = AllItems - SelItems; break;
 }

 // Make the progress dialog box
 WINASSERT (hwndProgressDlg == NULL);
 CreateDialogParam (hInst, MAKEINTRESOURCE (IDD_PROGRESSDLG) , hwndMain , (DLGPROC) ProgressProc , EXTRACT_PROGRESS);

 CResString Status (hInst , IDS_EXTRACTINGFILES);
 SetWindowText (hwndTextCtrl , Status);

 // We scan through every single item. For each one, we check to see if it is
 // one that we want to process (based on the method the user wants to use to
 // extract items). If it is one we want we process it and then update the
 // progress control. Otherwise, we skip this loop and go on to the next one.
 int ThisItem = 0;
 int SearchIndex = 0;
 WPARAM Pct;
 LV_ITEM lvi;
 while (SearchIndex < AllItems) {
     // Get information about this item.
     lvi.iItem = SearchIndex;
     lvi.iSubItem = 0;
     lvi.mask = LVIF_IMAGE | LVIF_PARAM | LVIF_STATE;
     lvi.stateMask = LVNI_SELECTED;
     ListView_GetItem (hwndListview , &lvi);

     // Now, we determine if this is one that we want to process or not. For
     // any we don't want to process, we increment SearchIndex and then
     // continue the loop.
     switch (ExtractProcess) {
         // We want everything, this is fine
         case 0: 
             break;

         // We only want this item if it is selected.
         case 1:
             if (lvi.state != LVNI_SELECTED) {
                 SearchIndex++;
                 continue;
             }
             break;

         // We only want this item if it is not selected
         case 2:
             if (lvi.state == LVNI_SELECTED) {
                 SearchIndex++;
                 continue;
             }
             break;
     }

     // If we get here we want to process this bad boy. On error get rid of
     // the progress dialog box and return.
     if (!ExtractEntry (lvi , FALSE , NULL)) {
         // Get rid of the progress dialog
         DestroyWindow (hwndProgressDlg);
         return;
     }

     if (ThisItem == TotalItems)
         Pct = 100;
     else
         Pct = (WPARAM) (((float) ThisItem / (float) TotalItems) * 100.0);

     SearchIndex++;
     ThisItem++;
     SendMessage (hwndProgressCtrl , PBM_SETPOS , Pct , 0);

     // If we have now processed the last item, break out of the loop.
     if (ThisItem == TotalItems)
         break;
 }

 // Get rid of the progress dialog
 DestroyWindow (hwndProgressDlg);
}

/****************************************************************************/

