/****************************************************************************/

#include "nfi.h"

/*****************************************************************************
 * Name         - CenterWindow
 * Purpose      - This centers the child window given over the other window
 *                given. Used to put the dialogs used by the program into
 *                a more desirable position.
 * Arguments    - hwndChild: The window to center
 *                hwndParent: The parent to center the child over
 * Return Value - TRUE on Successes, FALSE on Failure
 ****************************************************************************/
BOOL CenterWindow (HWND hwndChild , HWND hwndParent)
{
 WINASSERT (hwndChild);
 WINASSERT (hwndParent);

 RECT    rcChild , rcParent;
 int     cxChild , cyChild , cxParent , cyParent;
 int     cxScreen , cyScreen , xNew , yNew;
 HDC     hdc;

 // Get the Height and Width of the child window
 GetWindowRect (hwndChild , &rcChild);
 cxChild = rcChild.right - rcChild.left;
 cyChild = rcChild.bottom - rcChild.top;

 // Get the Height and Width of the parent window
 GetWindowRect (hwndParent , &rcParent);
 cxParent = rcParent.right - rcParent.left;
 cyParent = rcParent.bottom - rcParent.top;

 // Get the display limits
 hdc = GetDC (hwndChild);
 cxScreen = GetDeviceCaps (hdc , HORZRES);
 cyScreen = GetDeviceCaps (hdc , VERTRES);
 ReleaseDC (hwndChild , hdc);

 // Calculate new X position, then adjust for screen
 xNew = rcParent.left + ((cxParent - cxChild) / 2);
 if (xNew < 0) {
     xNew = 0;
 } else
     if ((xNew + cxChild) > cxScreen) {
         xNew = cxScreen - cxChild;
     }

 // Calculate new Y position, then adjust for screen
 yNew = rcParent.top  + ((cyParent - cyChild) / 2);
 if (yNew < 0) {
     yNew = 0;
 } else
     if ((yNew + cyChild) > cyScreen) {
         yNew = cyScreen - cyChild;
     }

 // Set it, and return
 return SetWindowPos (hwndChild , NULL , xNew , yNew , 0 , 0 ,
                      SWP_NOSIZE | SWP_NOZORDER);
}

/*****************************************************************************
 * Name         - MergeFilename
 * Purpose      - This function takes a caption string and a filename string
 *                and merges the two of them together. The resulting string is
 *                then used to alter the caption of the window passed in. If
 *                the filename string is a null string, the string resource
 *                IDS_UNTITLED will be loaded and used instead.
 * Arguments    - hWnd: Handle of the window whose caption is being altered
 *                Caption: Normal caption text
 *                Filename: The filename to merge with the caption
 * Return Value - None
 ****************************************************************************/
void MergeFilename (HWND hWnd , char *Caption , char *Filename)
{
 WINASSERT (hWnd);
 WINASSERT (Caption);
 WINASSERT (Filename);

 char AllBuffer[MAX_PATH + 256];
 CResString UntitledStr (hInst , IDS_UNTITLED);

 strcpy (AllBuffer , Caption);
 strcat (AllBuffer , " - ");

 if (Filename[0] == '\0')
     strcat (AllBuffer , UntitledStr);
 else
     strcat (AllBuffer , Filename);

 SetWindowText (hWnd , AllBuffer);
}

/*****************************************************************************
 * Name         - IdleProcessing
 * Purpose      - This function is repeatedly called to do work whenever the
 *                main message loop doesn't have any pending messages in it.
 *                We use this to do command enabling for the toolbars. Note
 *                that we don't need to command enable for the menubars
 *                because we get told before they are displayed, so it is best
 *                to do it then. For the toolbars, however, they are always
 *                visisble so we try to keep them as synced as possible with
 *                the program state at all times.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void IdleProcessing (void)
{
 // This tracks wether or not there where files selected the last time through
 // the loop. It it initialized to FALSE, since when the program is first
 // started, there aren't any selected items, and also, if a file is loaded
 // in at startup, no item will be selected.
 static BOOL FilesWhereSelected = FALSE;
 static BOOL FileWasDirty=FALSE;
 static LONG PreviousFileCount=-1;

 LONG FilesCurrentlySelected = ListView_GetSelectedCount (hwndListview);

 // Now, determine if there are any files selected at the moment
 BOOL FilesAreSelected = (FilesCurrentlySelected > 0);

 // Next. If there are files selected now but there wasn't before, OR, there
 // aren't now but there where before, the selection state has changed. If
 // that happens, we need to re-enable the buttons.
 if (FilesAreSelected != FilesWhereSelected) {
     // We only get in here if the selection state has changed. If
     // FilesAreSelected is TRUE, then it wasn't before so we need to enable
     // buttons. Otherwise, we disable.
     SendMessage (hwndToolbar , TB_ENABLEBUTTON , IDM_DELETEFILES ,
                  MAKELPARAM (FilesAreSelected , 0));
     SendMessage (hwndToolbar , TB_ENABLEBUTTON , IDM_MOVEFILES ,
                  MAKELPARAM (FilesAreSelected , 0));
     SendMessage (hwndToolbar , TB_ENABLEBUTTON , IDM_VIEWFILES ,
                  MAKELPARAM (FilesAreSelected , 0));

     // Now that we're done, change the static flag to the current situation
     // for next time through the loop
     FilesWhereSelected = FilesAreSelected;
 }

 // If the state of the file has changed since last time, change the enable
 // state of the save button.
 if (FileWasDirty != FileIsDirty) {
     SendMessage (hwndToolbar , TB_ENABLEBUTTON , IDM_FILESAVE ,
                  MAKELPARAM (FileIsDirty , 0));
     FileWasDirty = FileIsDirty;
 }

 // If the count of selected files has changed since last time, update the
 // status bar.
 if (FilesCurrentlySelected != FileSelectedCount) {
     FileSelectedCount = FilesCurrentlySelected;
     UpdateSelectedCount ();
 }

 // Has the count of files changed since last time?
 if (PreviousFileCount != FilesInNFI + FilesToAdd) {
     // Enable/Disable the "extract" button depending on if there are actually
     // any files that can be extracted (selected or otherwise).
     SendMessage (hwndToolbar , TB_ENABLEBUTTON , IDM_EXTRACTFILES ,
                  MAKELPARAM (FilesInNFI + FilesToAdd != 0 , 0));
     PreviousFileCount = FilesInNFI + FilesToAdd;
 }
}

/*****************************************************************************
 * Name         - PromptToSave
 * Purpose      - Prompts the user with a YES/NO/CANCEL message box and then
 *                returns the ID of the button the user pressed in the box.
 *                The Title string and the Prompt string are passed in as the
 *                resource string ID's of the strings to use.
 * Arguments    - TitleId: String ID of the title of the message box
 *                PromptId: String ID of the prompt for the message box
 * Return Value - IDYES, IDNO or IDCANCEL
 ****************************************************************************/
int PromptToSave (int TitleId , int PromptId)
{
 CResString Title (hInst , TitleId);
 CResString Prompt (hInst , PromptId);

 return MessageBox (hwndMain , Prompt , Title , MB_YESNOCANCEL | MB_ICONQUESTION);
}

/*****************************************************************************
 * Name         - PumpMessages
 * Purpose      - This function is called during lengthy operations to keep
 *                the program responsive. It checks to see if there are any
 *                messages for the application, and if so it gets them and
 *                dispatches them. It does not dispatch messages that might
 *                result in the program terminating (WM_CLOSE, WM_DESTROY)
 *                since we don't want the user to be able to quit while the
 *                program is doing something. Note, however, that the message
 *                is removed from the queue and then discarded.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void PumpMessages (void)
{
 static MSG Msg;

 // See if there is a message
 if (PeekMessage (&Msg , NULL , 0 , 0 , PM_NOREMOVE)) {
     // If GetMessage returns 0, it means that a WM_QUIT message was
     // received. If that is the case, just return.
     if (!GetMessage (&Msg , NULL , 0 , 0))
         return;

     if (Msg.message == WM_QUERYENDSESSION)
         return;

     // Dispatch it
     if (!TranslateAccelerator (hwndMain , hAccelTable , &Msg)) {
         TranslateMessage (&Msg);
         DispatchMessage (&Msg);
     }
 }
}

/*****************************************************************************
 * Name         - AddPathToComboBox
 * Purpose      - This recursive function traverses the binary tree that
 *                holds all of the unique paths in the current NFI file. For
 *                every node, it adds the directory to the listbox portion of
 *                the combobox whose handle it is given.
 * Arguments    - Node: Node containing the path to add
 *                Control: Window handle of the combobox to add to
 * Return Value - None
 ****************************************************************************/
void AddPathToComboBox (PathInfo *Node , HWND Control)
{
 WINASSERT (Control);

 if (Node != NULL) {
     AddPathToComboBox (Node->Left , Control);
     WINASSERT (Node->Path);
     SendMessage (Control , CB_ADDSTRING ,
                  (WPARAM) 0 ,
                  (LPARAM) Node->Path);
     AddPathToComboBox (Node->Right , Control);
 }
}

/*****************************************************************************
 * Name         - UnlinkDirEntry
 * Purpose      - This function takes a pointer to the given directory entry
 *                in the doubly linked list of loaded NFI entries and removes
 *                it from the list without destroying all of the other links.
 *                once that is done, the entry is deleted.
 * Arguments    - Entry: The directory entry to remove
 * Return Value - None
 ****************************************************************************/
void UnlinkDirEntry (NFIDirEntry *Entry)
{
 WINASSERT (Entry);

 // Ok. What are the possible things we have to deal with in removing this
 // item.
 // 1) This is the only item in the list
 // 2) This is the first item in the list
 // 3) This is the last item in the list
 // 4) It's in the middle of the list
 // Handle each case seperately.

 // Is this is the only item in the list?
 if (Entry->Prev == NULL && Entry->Next == NULL) {
     // This is the only thing in the list. Set the head to NULL and the
     // removal process is done.
     DirectoryRoot = NULL;
 } else
     // Is this the last item in the list?
     if (Entry->Next == NULL) {
         // This is the last thing in the list. We set our previous item's
         // next to NULL, making it the new last item in the list.
         Entry->Prev->Next = NULL;
     } else
         // Is this the first item in the list?
         if (Entry->Prev == NULL) {
             // This is the first item in the list. We set the head to our
             // next, making it the new head, then set our next's previous to
             // NULL, so that it knows it's first.
             DirectoryRoot = Entry->Next;
             Entry->Next->Prev = NULL;
         } else {
             // This item must be in the middle of the list. We set our
             // previous item's next, to our next. Then, we set our next
             // item's previous to our previous. That overlinks the nodes
             // on either side of us and gets this node out of the loop.
             Entry->Prev->Next = Entry->Next;
             Entry->Next->Prev = Entry->Prev;
         }

 // Now we can safely delete this node.
 char *Temp = (char *) Entry;

 // Nuke the memory for this entry
 delete[] Temp;
}

/*****************************************************************************
 * Name         - UnlinkAddEntry
 * Purpose      - This function takes a pointer to the given directory entry
 *                in the doubly linked list of waiting to be added files and
 *                removes it from the list without destroying all of the other
 *                links. once that is done, the entry is deleted.
 * Arguments    - Entry: The entry to remove
 * Return Value - None
 ****************************************************************************/
void UnlinkAddEntry (AddedFile *Entry)
{
 WINASSERT (Entry);

 // Ok. What are the possible things we have to deal with in removing this
 // item.
 // 1) This is the only item in the list
 // 2) This is the first item in the list
 // 3) This is the last item in the list
 // 4) It's in the middle of the list
 // Handle each case seperately.

 // Is this is the only item in the list?
 if (Entry->Prev == NULL && Entry->Next == NULL) {
     // This is the only thing in the list. Set the head to NULL and the
     // removal process is done.
     AddedFileList = NULL;
 } else
     // Is this the last item in the list?
     if (Entry->Next == NULL) {
         // This is the last thing in the list. We set our previous item's
         // next to NULL, making it the new last item in the list.
         Entry->Prev->Next = NULL;
     } else
         // Is this the first item in the list?
         if (Entry->Prev == NULL) {
             // This is the first item in the list. We set the head to our
             // next, making it the new head, then set our next's previous to
             // NULL, so that it knows it's first.
             AddedFileList = Entry->Next;
             Entry->Next->Prev = NULL;
         } else {
             // This item must be in the middle of the list. We set our
             // previous item's next, to our next. Then, we set our next
             // item's previous to our previous. That overlinks the nodes
             // on either side of us and gets this node out of the loop.
             Entry->Prev->Next = Entry->Next;
             Entry->Next->Prev = Entry->Prev;
         }

 // Now we can safely delete this node.
 char *Temp = (char *) Entry;

 // Nuke the memory for this entry
 delete[] Entry->AddFrom;
 delete[] Entry->InternalName;
 delete[] Entry->InternalPath;
 delete[] Temp;
}

/*****************************************************************************
 * Name         - GetHelpFileName
 * Purpose      - This constructs a fully qualified help file name and
 *                puts it into the given buffer. The base name is loaded from
 *                the string resource table of the application. Then it 
 *                determines the directory the executable is in and adds that
 *                to obtain the full pathname of the helpfile.
 * Arguments    - Filename: Buffer to put the filename into
 * Return Value - None
 ****************************************************************************/
void GetHelpFileName (char *Filename)
{
 WINASSERT (Filename);

 char ExeName[MAX_PATH];
 char Drive[_MAX_DRIVE];
 char Path[MAX_PATH];

 // Get the base name.
 CResString HelpFile (hInst , IDS_HELPFILE);

 // Get the EXE file name.
 if (!GetModuleFileName (hInst , ExeName , sizeof (ExeName))) {
     strcpy (Filename , HelpFile);
     return;
 }

 // Split the name apart then construct a fully qualified path name using
 // the loaded help file name.
 _splitpath (ExeName , Drive , Path , NULL , NULL);
 strcpy (Filename , Drive);
 strcat (Filename , Path);
 strcat (Filename , HelpFile);
}

/****************************************************************************/

