// An NFI file is an archived file. It has it's own internal file system,
// comparible with a Quake .PAK file, allowing multiple files to be stored
// inside the main file, and randomly accessed.
//
// The NFI file starts with a Header which is used to identify the file as
// an NFI file, as well as giving some general information about the file.
// The header contains a file offset to where the directory information for
// the file is stored.
//
// The directory information consists of a count of the number of files in
// the NFI file, and is followed by one directory entry for every file in
// the directory.
//
// A directory entry consists of the name of a file, it's length in bytes,
// and the offset from the start of the NFI where it's data begins. Using
// this information, a program can seek to the file's location in the NFI
// file and read from it as if it was reading the file off the disk.
//
// Currently, the NFI format does not include any CRC or compress/encryption
// options, although the structure of the file has been designed to allow
// the addition of such options in the future.

// This defines how long the filename in an NFI can be (including any
// directory information). This takes the null into account
#define NFI_NAMELEN 260

// This defines the NFI version information. This will be bumped whenever the
// file format changes. Any reading program can use this information to
// determine the format of the file, or just summarily reject the file
#define NFI_VERSION 1

// This is the string which identifies a file as an NFI File. Note that a
// string of "NFI" is actually 4 characters, there is an implied NULL
// character on the end. That NULL is part of the header ID
#define NFI_HEADERID "NFI"

// The definition of the header of an NFI File. This is read in when the
// file is opened to get information about the file.
typedef struct {
    char          IDTag[4];  // Identifying marker. Should be NFI_HEADERID
    unsigned char Version;   // Identifies the version of the NFI File
    unsigned long Reserved;  // Space reserved for future versions
    unsigned long DirOffset; // Offset of the directory from the start of the file
} NFIHeader;

// The definition of the directory header of an NFI File. The location of this
// information is given in the file header. It can be anywhere in the file.
// Normally, the directory header and the directory information itself is
// stored directly after the header by the NFI Tool. However, code should
// not assume that.
typedef struct {
    unsigned long NumEntries; // Number of entries in the directory
    unsigned long Reserved;   // Reserved space for future versions
} NFIDirHeader;

// The definition of a directory entry in an NFI File. It gives the information
// about what the file is named, where in the NFI file it is stored, and how
// long it actually is. The last two members of this structure, Next and Prev,
// are not stored in the actual NFI File. They is used internally by the code,
// to store all of the directory information in either A) a linked list (single
// or double, or B) a binary tree. The reading (or writing) code can use any
// of these methods it prefers, depending on the ease of programming or the
// speed the application requires. The Other field is used in a similar matter.
// It is not stored in the file, but is included in this data structure so that
// the program can attach some other data to a node in memory.
typedef struct _NFIDirEntryTag {
    char            Name[NFI_NAMELEN];     // Name of the file
    unsigned long   Offset;                // Offset into the file
    unsigned long   Length;                // Length of the file in bytes
    unsigned long   Reserved[2];           // Reserved space for future versions
    _NFIDirEntryTag *Next;                 // Used internally only.
    _NFIDirEntryTag *Prev;                 // Used internally only.
    void            *Other;                // Used internally only.
} NFIDirEntry;
