/****************************************************************************/

#include "nfi.h"

/*****************************************************************************
 * Name         - InitializeStatusBar
 * Purpose      - This function initializes the statusbar panes by determining
 *                how big each should be. This is called when the statusbar
 *                is first created, and whenever the statusbar changes sizes.
 * Arguments    - hwndParent: Handle to the parent of the statusbar.
 * Return Value - None
 ****************************************************************************/
void InitializeStatusbar (HWND hwndParent)
{
 WINASSERT (hwndParent);
 const cSpaceInBetween = 8;
 char  Buffer[1024];
 int   ptArray[4] = {0 , 0 , 0 , 0};
 SIZE  Size;
 RECT  Rect;
 HDC   hDC;

 // Get a Device context so we can get some metrics information. Also, get
 // the bounding rectangle of the parent, so we can tell how wide we are.
 hDC = GetDC (hwndParent);
 GetClientRect (hwndParent , &Rect);

 // We now determine where the right side of every status pane begins. For the
 // last pane, this is the edge of the window. For the other panes, it depends
 // on what is in the pane to it's right. So, for example, to determine where
 // the third pane should end, we find out how big the pane to it's right is,
 // then subtract that from where that pane ends to see where this one ends.
 // We also take into account a little space to make sure things are separated
 // ok

 // The last statusbar pane always ends at the right edge of the window.
 ptArray[STATBAR_PANE4] = Rect.right;

 CResString ModStr (hInst , IDS_CHANGED);
 if (GetTextExtentPoint32 (hDC, ModStr , strlen (ModStr) , &Size)) {
     ptArray[STATBAR_PANE3] = ptArray[STATBAR_PANE4] -
                              (Size.cx) - cSpaceInBetween;
 }

 CResString SelectStr (hInst , IDS_FILESELECTCOUNT);
 wsprintf (Buffer , SelectStr , 10000);
 if (GetTextExtentPoint32 (hDC, Buffer , strlen (Buffer) , &Size)) {
     ptArray[STATBAR_PANE2] = ptArray[STATBAR_PANE3] - (Size.cx) -
                              cSpaceInBetween;
 }

 CResString CountStr (hInst , IDS_FILECOUNTWITHADD);
 wsprintf (Buffer , CountStr , 10000 , 1000);
 if (GetTextExtentPoint32 (hDC, Buffer , strlen (Buffer) , &Size)) {
     ptArray[STATBAR_PANE1] = ptArray[STATBAR_PANE2] - (Size.cx) -
                              cSpaceInBetween;
 }

 // We don't need the DC anymore.
 ReleaseDC (hwndParent , hDC);

 // Give the newly calculated information to the status bar so it can resize
 // the panes.
 WINASSERT (hwndStatusbar);
 SendMessage (hwndStatusbar , SB_SETPARTS ,
              sizeof (ptArray) / sizeof (ptArray[0]) ,
              (LPARAM) (LPINT) ptArray);

 // Now update the contents of the status bar
 // Set the contents of pane 1 to the name of our program
 CResString VerStr (hInst , IDS_VERSIONSTRING);
 UpdateStatusbar (VerStr , STATBAR_PANE1 , 0);

 // Set pane 2 to the number of files in the NFI File
 UpdateFileCount ();

 // Set pane 3 to the number of selected files in the NFI File
 UpdateSelectedCount ();

 // Set pane 4 to the current file status
 SetFileStatus (FileIsDirty);
}

/*****************************************************************************
 * Name         - CreateStatusBar
 * Purpose      - This creates the statusbar
 * Arguments    - hwndParent: Parent window of the statusbar
 * Return Value - TRUE if the statusbar was created
 *                FALSE if some error occured
 ****************************************************************************/
BOOL CreateStatusbar (HWND hwndParent)
{
 WINASSERT (hwndParent);
 WINASSERT (hwndStatusbar == NULL);

 // Create the status bar.
 hwndStatusbar = CreateStatusWindow (WS_CHILD | WS_VISIBLE | WS_BORDER,
                                     "" , hwndParent,
                                     ID_STATUSBAR);

 // Was it created?
 if (hwndStatusbar) {
     // Set up the status bar panes
     InitializeStatusbar (hwndParent);
     return TRUE;
 }

 return FALSE;
}

/*****************************************************************************
 * Name         - UpdateStatusbar
 * Purpose      - This sets a specified pane of the statusbar to the given
 *                text.
 * Arguments    - String: The string to put into the status bar
 *                PartNumber: The pane to insert the string into
 *                DisplayFlags: Display flags that control how the text looks
 * Return Value - None
 ****************************************************************************/
void UpdateStatusbar (char *String , WORD PartNumber , WORD DisplayFlags)
{
 WINASSERT (String);
 WINASSERT (PartNumber <= STATBAR_PANE4 || PartNumber == STATBAR_PANEALL);
 WINASSERT (hwndStatusbar);
 SendMessage (hwndStatusbar , SB_SETTEXT , PartNumber | DisplayFlags ,
              (LPARAM) String);
}

/*****************************************************************************
 * Name         - UpdateFileCount
 * Purpose      - This updates the status bar pane with the number of files
 *                that are currently in the NFI file, as well as the number
 *                of files that are waiting to be added
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void UpdateFileCount (void)
{
 char Buffer[1024];

 // We use a different string depending on if there are files waiting to be
 // added to the NFI file or not.
 if (FilesToAdd > 0) {
     CResString Str (hInst , IDS_FILECOUNTWITHADD);
     wsprintf (Buffer , Str , FilesInNFI , FilesToAdd);
 } else {
     CResString Str (hInst , IDS_FILECOUNTNOADD);
     wsprintf (Buffer , Str , FilesInNFI);
 }

 // Update the status bar
 UpdateStatusbar (Buffer , STATBAR_PANE2 , 0);
}

/*****************************************************************************
 * Name         - UpdateSelectedCount
 * Purpose      - This updates the status bar pane that displays the count of
 *                selected files.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void UpdateSelectedCount (void)
{
 CResString Str (hInst , IDS_FILESELECTCOUNT);
 char Buffer[1024];

 wsprintf (Buffer , Str , FileSelectedCount);
 UpdateStatusbar (Buffer , STATBAR_PANE3 , 0);
}

/*****************************************************************************
 * Name         - SetFileStatus
 * Purpose      - Changes the Dirty status of the file and then updates the
 *                statusbar pane that tracks it
 * Arguments    - Status: New file status
 * Return Value - None
 ****************************************************************************/
void SetFileStatus (BOOL Status)
{
 FileIsDirty = Status;

 if (FileIsDirty) {
     CResString Str (hInst , IDS_CHANGED);
     UpdateStatusbar (Str , STATBAR_PANE4 , 0);
 } else {
     CResString Str (hInst , IDS_NOTCHANGED);
     UpdateStatusbar (Str , STATBAR_PANE4 , 0);
 }
}

/****************************************************************************/

