/****************************************************************************/

// This typedef is used to refer to a data handle used with the handle
// functions of the data loader.
typedef int DATAHANDLE;

// These data types are specific to the NFI Handling code. The sizes were mirrored
// from the NFI Tool code that was most current at the time that it was written.
// That means that this code can only read older files until someone specifically
// changes it (assuming the NFI format ever changes).
#define NFI_NAMELEN  260
#define NFI_VERSION  1
#define NFI_HEADERID "NFI"

// The definition of an NFI File header
typedef struct {
    char  ID[4];
    char  Version;
    ULONG Reserved;
    ULONG DirOffs;
} NFIFileHeader;

// The definition of an NFI directory header
typedef struct {
    ULONG DirSize;
    ULONG Reserved;
} NFIDirHeader;

// The definition of an NFI directory entry
typedef struct {
    char  Name[NFI_NAMELEN];
    ULONG Offset;
    ULONG Length;
    char  Reserved[8];
} NFIDirEntry;

// The definition of an actual NFI File. Contains a file header, a directory
// header, and a dynamically sized array that contains the actual directory
// of the file. There is also a next pointer because all of the loaded NFI
// files are stored in a linked list.
typedef struct TAGNFIFile {
    char          Name[MAX_PATH];
    NFIFileHeader FileHeader;
    NFIDirHeader  DirHeader;
    NFIDirEntry   *Directory;
    HANDLE        FileHandle;
    TAGNFIFile *Next;
} NFIFile;

/****************************************************************************/

// This data loader is used to allow a program to transparently read from
// regular files in the filesystem, or from an archived NFI format file,
// without actually knowing where the data is coming from. This allows for
// operation like in Quake, where all data files may be placed in a PAK file
// or read from the hard drive.
//
// In order to achieve this level of absraction, the Loader has it's own
// routines for the opening and reading of a data file. All filenames must
// be in the FSYS Format (which is described below). The return is an
// object of type DATAHANDLE, which is used in all file operations to
// identify the file that is open.
//
// Files that exist on the hard drive have preference over files contained
// in an NFI archive. This allows the overriding of files in the archive
// by putting a revised version on disk.
//
// When the loader is initialized, it must be passed a string which contains
// the fully qualified executable filename. From this it extracts the pathname
// that the executable is in. After appending the contents of the DATA_SUBDIR
// define, the resulting path is considered the local data path. This path
// on the actual hard drive corrosponds to the single / path in an FSYS
// filename. This means that if the executable is in C:\Test and DATA_SUBDIR
// is set to "\Data", an FSYS filename of "/gfx/filename.pcx" will first try
// to be loaded from C:\Test\Data\gfx\filename.pcx. If that doesn't work,
// an NFI entry corrosponding to "/gfx/filename.pcx" will be loaded instead.
//
// As is probably obvious, an FSYS style filename is one which contains
// a filename and an optional path. Paths are specified with a / character
// and must always begin with a / character, which represents the base data
// directory. Filenames in the loader are not case sensitive.
//
// Sequence of calls to use this is:
//   - InitializeDataLoader ()
//   - OpenNFIFiles ()
//       - Open/Read/Close file functions, starting with FSYS
//   - ShutdownDataLoader ()

/****************************************************************************/

// This defines the maximum number of files the loader will allow open at one
// time. This applies most generally to the files opened from a larger
// packed NFI file, since the operating system itself will probably not allow
// this many files to be open simultaneously.
#define MAX_DATA_HANDLES 256

// THis defines what subdirectory (if any) should be assumed to contain the
// data structures if the file is being loaded from the disk and not the file
// system. This is added onto the directory that the EXE started in, to obtain
// the local directory that should be considered the "root" of the data. Make
// this a NULL string to say that data files may possibly reside in the same
// directory as the .EXE file.
#define DATA_SUBDIR "\\DATA"

/****************************************************************************/

// This is the structure which reprents a handle as used by the loader code.
// An array of such structs is how the loader hands out handles. RealFile is
// a windows filesystem handle to the actual file the datahandle represents.
// For an NFI file it;s the handle to the NFI file the file resides inside.
// For a local file, it's a handle to the actual file. The IsPacked boolean
// is TRUE if the file is inside an NFI and false if not. StartPos is the
// file offset inside the NFI file that the data for this file resides in, or
// 0 if it's a local file. Length is always the length in bytes of the file.
typedef struct {
    HANDLE RealFile;
    BOOL   IsPacked;
    ULONG  StartPos;
    ULONG  Length;
    ULONG  Position;
} InternalHandle;

/****************************************************************************/

// This variable is what the loading routines consider to be the base directory
// of all data operatoins.
static char RootDir[MAX_PATH] = "/";

// This holds the list of handles, which are mapped to the data handles that
// the functions use. This allows the calling code to transparently use the
// same handle value for a real file or an NFI Component file.
static InternalHandle HandleList[MAX_DATA_HANDLES];

// This holds the list of loaded NFI Files that are ready to go.
NFIFile *NFIList = NULL;
static ULONG   NFICount = 0;

/*****************************************************************************
 * Name         - ReadNFIFileHeader
 * Purpose      - Reads the NFI File Header from an NFI file and stores that
 *                header in the given struct. THe fuction assumes that the file
 *                handle in the NFIFile struct passed in is valid and the file
 *                to load from, and also that it is at the start of the file.
 *                If the function fails it will close the file before returning.
 * Arguments    - Info - NFIFile structure to read information into
 * Return Value - TRUE on Success, FALSE on failure
 ****************************************************************************/
BOOL ReadNFIFileHeader (NFIFile &Info)
{
 DWORD Read;
 // Try to read in the various parts of the main header one at a time. All of
 // the reads are enclosed in a try block. If there is any error reading, an
 // exception will be thrown.
 try {
     // Get the header ID
     if (!ReadFile (Info.FileHandle , Info.FileHeader.ID ,
               sizeof (Info.FileHeader.ID) , &Read , NULL) ||
	       Read != sizeof (Info.FileHeader.ID))
         throw 1;

     // Get the file version
     if (!ReadFile (Info.FileHandle , &Info.FileHeader.Version ,
               sizeof (Info.FileHeader.Version) , &Read , NULL) ||
	       Read != sizeof (Info.FileHeader.Version))
         throw 1;

     // Get the reserved info
     if (!ReadFile (Info.FileHandle , &Info.FileHeader.Reserved ,
               sizeof (Info.FileHeader.Reserved) , &Read , NULL) ||
	       Read != sizeof (Info.FileHeader.Reserved))
         throw 1;

     // Get the directory offset
     if (!ReadFile (Info.FileHandle , &Info.FileHeader.DirOffs ,
               sizeof (Info.FileHeader.DirOffs) , &Read , NULL) ||
	       Read != sizeof (Info.FileHeader.DirOffs))
         throw 1;

     // If the header ID and file version aren't right, toss an error.
     if (Info.FileHeader.Version != NFI_VERSION)
         throw 1;

     if (strcmpi (Info.FileHeader.ID , NFI_HEADERID) != 0)
         throw 1;
 }

 // This catch block will catch any thrown exception. It closes the file handle
 // that was passed in and then returns false.
 catch (...) {
     CloseHandle (Info.FileHandle);
     return FALSE;
 }

 // Otherwise, if we are here, it worked
 return TRUE;
}

/*****************************************************************************
 * Name         - ReadNFIDirHeader
 * Purpose      - Reads the NFI Directory Header from an NFI file and stores that
 *                header in the given struct. THe fuction assumes that the file
 *                handle in the NFIFile struct passed in is valid and the file
 *                to load from. The function seeks to the directory before reading.
 *                If the function fails it will close the file before returning.
 *                The main header must have been read into the struct before
 *                calling this because the directory offset is contained there.
 * Arguments    - Info - NFIFile structure to read information into
 * Return Value - TRUE on Success, FALSE on failure
 ****************************************************************************/
BOOL ReadNFIDirHeader (NFIFile &Info)
{
 DWORD Read;
 // Try to read in the various parts of the main header one at a time. All of
 // the reads are enclosed in a try block. If there is any error reading, an
 // exception will be thrown.
 try {
     // Try to seek to the directory position.
     if (SetFilePointer (Info.FileHandle , Info.FileHeader.DirOffs , NULL , FILE_BEGIN)
         == 0xFFFFFFFF)
         throw 1;

     // Get the directory size
     if (!ReadFile (Info.FileHandle , &Info.DirHeader.DirSize ,
               sizeof (Info.DirHeader.DirSize) , &Read , NULL) ||
	       Read != sizeof (Info.DirHeader.DirSize))
         throw 1;

     // Get the reserved info
     if (!ReadFile (Info.FileHandle , &Info.DirHeader.Reserved ,
               sizeof (Info.DirHeader.Reserved) , &Read , NULL) ||
	       Read != sizeof (Info.DirHeader.Reserved))
         throw 1;
 }

 // This catch block will catch any thrown exception. It closes the file handle
 // that was passed in and then returns false.
 catch (...) {
     CloseHandle (Info.FileHandle);
     return FALSE;
 }

 // Otherwise, if we are here, it worked
 return TRUE;
}

/*****************************************************************************
 * Name         - ReadNFIDirEntries
 * Purpose      - After reading in the main header and directory header, this
 *                function will read in the entire directory of the NFI File.
 *                It generates a dynamic array to store the entries in. This
 *                code assumes that the file pointer is where the ReadNFIDirHeader
 *                function left it, since that is the position of the first
 *                entry. The file will be closed and the dynamic array released
 *                if there is an error.
 * Arguments    - Info - NFIFile structure to read information into
 * Return Value - TRUE on Success, FALSE on failure
 ****************************************************************************/
BOOL ReadNFIDirEntries (NFIFile &Info)
{
 DWORD Read;

 // The file postion is in just the rght place to start reading. First thing to
 // do is create the dynamic array. We don't need to check this for failure
 // because the new handler that WinLib installs will toss an exception if we
 // run out of memory.
 Info.Directory = new NFIDirEntry[Info.DirHeader.DirSize];

 // This here loop that reads the entries are in a try block. Any file error will
 // throw an exception that is later caught to clean things up.
 try {
     // Read all of the entries
     for (ULONG Index = 0 ; Index < Info.DirHeader.DirSize ; Index++) {
         // Get the filename
         if (!ReadFile (Info.FileHandle , Info.Directory[Index].Name ,
                   sizeof (Info.Directory[Index].Name) , &Read , NULL) ||
                   Read != sizeof (Info.Directory[Index].Name))
             throw 1;

         // Get the offset
         if (!ReadFile (Info.FileHandle , &Info.Directory[Index].Offset ,
                   sizeof (Info.Directory[Index].Offset) , &Read , NULL) ||
                   Read != sizeof (Info.Directory[Index].Offset))
             throw 1;

         // Get the length
         if (!ReadFile (Info.FileHandle , &Info.Directory[Index].Length ,
                   sizeof (Info.Directory[Index].Length) , &Read , NULL) ||
                   Read != sizeof (Info.Directory[Index].Length))
             throw 1;

         // Get the reserved info
         if (!ReadFile (Info.FileHandle , Info.Directory[Index].Reserved ,
                   sizeof (Info.Directory[Index].Reserved) , &Read , NULL) ||
                   Read != sizeof (Info.Directory[Index].Reserved))
             throw 1;
     }
 }

 // This catch block will catch any thrown exception. It closes the file handle
 // that was passed in and then returns false.
 catch (...) {
     delete[] Info.Directory;
     Info.Directory = NULL;
     CloseHandle (Info.FileHandle);
     return FALSE;
 }

 // Otherwise, if we are here, it worked
 return TRUE;
}

/*****************************************************************************
 * Name         - GetNFIInfo
 * Purpose      - Called by the NFI Startup routines when it detects an NFI
 *                file. This tries to read in all of the directory and file
 *                information for that file.
 * Arguments    - FullName: The fully qualified name of the NFI file.
 *                BareName: The NFI Filename without directory info.
 * Return Value - If everything went OK, the return value is a pointer to a
 *                dynamically allocated NFIFile structure. Otherwise, NULL is
 *                returned.
 ****************************************************************************/
NFIFile *GetNFIInfo (char *FullName , char *BareName)
{
 NFIFile Temp;

 // Null out the entire structure.
 memset (&Temp , 0 , sizeof (Temp));

 // Firstly, set the name of the NFI File.
 strcpy (Temp.Name , BareName);

 // Try to open the file.
 Temp.FileHandle = CreateFile (FullName , GENERIC_READ , FILE_SHARE_READ ,
                               NULL , OPEN_EXISTING ,
			       FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS ,
                               NULL);

 // If it didn't work, return NULL
 if (Temp.FileHandle == INVALID_HANDLE_VALUE)
     return NULL;

 // All of the functions below here that read bits of the headers will close the
 // file if an error happens.

 // Try to get the header. Return NULL on failure.
 if (!ReadNFIFileHeader (Temp))
     return NULL;

 // Now try to get the Directory header. Return NULL on failure.
 if (!ReadNFIDirHeader (Temp))
     return NULL;

 // Try to read all of the directory entries. Return NULL on failure.
 if (!ReadNFIDirEntries (Temp))
     return NULL;

 // If we get here, everything must have worked out OK. In that case, create
 // a new dynamic node and copy this info over to it, then return it.
 NFIFile *Node = new NFIFile;
 memcpy (Node , &Temp , sizeof (NFIFile));
 Node->Next = NULL;

 return Node;
}

/*****************************************************************************
 * Name         - OpenNFIFiles
 * Purpose      - Scans the base data directory for all of the NFI Files that
 *                it contains. Every time it finds one it opens it and scans in
 *                the directory.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void OpenNFIFiles (void)
{
 // First thing we want to do is create a filespec to scan for.
 char SearchPath[MAX_PATH];
 char NFIFilename[MAX_PATH];

 strcpy (SearchPath , RootDir);
 strcat (SearchPath , "\\*.NFI");

 // Now Start a search and try to find the first item.
 HANDLE          SearchHandle;
 WIN32_FIND_DATA Found;

 // Try to find the first file
 SearchHandle = FindFirstFile (SearchPath , &Found);

 // If we didn't find anything then just leave, nothing more to do.
 if (SearchHandle == INVALID_HANDLE_VALUE)
     return;

 // Otherwise, handle the file we found, and keep looking till we're done.
 do {
     // Create a fully qualified filename of the NFI File
     strcpy (NFIFilename , RootDir);
     strcat (NFIFilename , "\\");
     strcat (NFIFilename , Found.cFileName);

     // Tell what file we found
     AddConsoleStringFmt ("Loading %s" , NFIFilename);

     // Try to load it and see if it works.
     NFIFile *Temp = GetNFIInfo (NFIFilename , Found.cFileName);

     // Try to open the file. If it worked add it to the NFI List and update
     // the NFI Count.
     if (Temp) {
         // If NFIList is NULL, this is the first entry. Otherwise, make this
         // the first entry and set it's next to the current head.
         if (!NFIList)
             NFIList = Temp;
         else {
             Temp->Next = NFIList;
             NFIList = Temp;
         }
         NFICount++;
     }
 } while (FindNextFile (SearchHandle , &Found));

 // Now we're done, close the search handle.
 FindClose (SearchHandle);
}

/*****************************************************************************
 * Name         - InitializeDataLoader
 * Purpose      - Starts up the data loader. This is a set of file routines
 *                that transparently load files from either a disk file or a
 *                packed file like an NFI file. The function stores the
 *                directory portion of the passed in exe filename and uses
 *                that directory as the location of the passed in PackedFile
 *                or the root directory portion of a local filename if the
 *                file is being loaded from disk.
 * Arguments    - ExePath: Fully qualified pathname of the program's
 *                         executable module.
 * Return Value - None
 ****************************************************************************/
void InitializeDataLoader (char *ExePath)
{
 WINASSERT (ExePath);

 // We need to determine the path portion of this EXE path. Because it is
 // fully qualified, all we have to do is scan backwards from the end of the
 // string until we locate a slash. We simply NULL that out and we're done.

 // Make a copy of the passed in string.
 strcpy (RootDir , ExePath);

 // Try to get a pointer to the last slash in the string. Check the other
 // type of slash if we don't find the normal one
 char *Temp = strrchr (RootDir , '\\');
 if (!Temp)
     Temp = strrchr (RootDir , '/');

 // If Temp is NULL when we get here there is something drastically wrong
 // so set the data root to an empty string.
 if (Temp == NULL)
     RootDir[0] = '\0';
 else
     // Otherwise temp refers to the last slash in the string. Make that a NULL
     // byte. The RootDir then becomes a fully qualified path name to the root
     // game directory which has no trailing slash (which is vitally important
     // to the loader routines).
     *Temp = '\0';

 // Now tack on the additional data portion.
 strcat (RootDir , DATA_SUBDIR);

 AddConsoleStringFmt ("Base data directory is %s\\" , RootDir);

 // Scan for and open all of the NFI files in the data directory.
 OpenNFIFiles ();

 // Output the number of NFI Files that were loaded
 if (!NFIList)
     AddConsoleString ("No NFI files found");
 else
     AddConsoleStringFmt ("%d NFI file%s loaded" , NFICount ,
                          NFICount > 1 ? "s" : "");

 // Scan the entire DataHandle directory and set all of the handles to
 // invalid handles so that later we can tell which entry is in use and which
 // is not.
 for (int Index = 0 ; Index < MAX_DATA_HANDLES ; Index++)
     HandleList[Index].RealFile = INVALID_HANDLE_VALUE;
}

/*****************************************************************************
 * Name         - ShutdownDataLoader
 * Purpose      - Does the work of closing down the data loader. Any open
 *                files will be closed.
 * Arguments    - None
 * Return Value - None
 ****************************************************************************/
void ShutdownDataLoader (void)
{
 // First of all, scan through the entire internal handle list. For every
 // entry we find that is a real file and not an NFI entry, close the file.
 for (int Index = 0 ; Index < MAX_DATA_HANDLES ; Index++) {
     // Is this entry in use
     if (HandleList[Index].RealFile != INVALID_HANDLE_VALUE)
         // If the file is not packed close it
         if (!HandleList[Index].IsPacked)
             CloseHandle (HandleList[Index].RealFile);
 }

 // Now close all of the open NFI Files and get rid of the directories in them.
 NFIFile *Temp = NFIList;
 while (Temp) {
     // Make a copy of the temp pointer, since that's what we are deleting.
     NFIList = Temp->Next;

     // Now we can delete Temp no problem
     CloseHandle (Temp->FileHandle);
     delete[] Temp->Directory;
     delete Temp;

     // Set temp to what the next entry is
     Temp = NFIList;
 }
 NFIList = NULL;
}

/*****************************************************************************
 * Name         - SearcnForNFIEntry
 * Purpose      - Scans all of the loaded NFI Files for a directory entry that
 *                matches the name that is passed in. If a match is found,
 *                Entry is set to the directory entry found, EntryFile is set
 *                to the NFIFile structure that the entry is contained in, and
 *                TRUE is returned. Otherwise, FALSE is returned and nothing is
 *                Altered.
 * Arguments    - Entry: Struct to save found directory entry in
 *                EntryFile: Struct to save NFI information of found entry
 *                Name: FSYS filename of entry to find
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL SearchForNFIEntry (NFIDirEntry &Entry , NFIFile &EntryFile , char *Name)
{
 // Scan through all of the loaded NFI File's.
 NFIFile *Search = NFIList;

 while (Search) {
     // Scan the entire directory of this NFI File to see if we can find one that
     // matches.
     for (ULONG Index = 0 ; Index < Search->DirHeader.DirSize ; Index++) {
         // Check if this entry matches.
         if (strcmpi (Name , Search->Directory[Index].Name) == 0) {
             // It matched. Set up the Entry and EntryFile structure's
             // and return TRUE.
             memcpy (&Entry , &Search->Directory[Index] , sizeof (NFIDirEntry));
             memcpy (&EntryFile , Search , sizeof (NFIFile));

             return TRUE;
         }
     }

     // Go on to the next NFI file in the list.
     Search = Search->Next;
 }

 // If we get here, nothing was found so return error.
 return FALSE;
}

/*****************************************************************************
 * Name         - FSYSOpenFile
 * Purpose      - Opens a file in the loader file system. The specified file
 *                is opened from one of the packed NFI files the file system
 *                is using, or from the local file system if no matching file
 *                is found.
 * Arguments    - Name: Filename to open
 * Return Value - -1 on error, else a data handle to use with other FSYS
 *                functions.
 ****************************************************************************/
DATAHANDLE FSYSOpenFile (char *Name)
{
 int Handle = -1;

 // Scan to find the first entry in the data handle list that isn't already
 // being used.
 for (int Index = 0 ; Index < MAX_DATA_HANDLES ; Index++)
     if (HandleList[Index].RealFile == INVALID_HANDLE_VALUE)
         break;

 // Once here, Index is the array index of the first open entry, or
 // MAX_DATA_HANDLES if they are all full. If that is the case we can't open
 // the file.
 if (Index == MAX_DATA_HANDLES)
     return -1;

 // Set our handle to the array index we will be using.
 Handle = Index;

 // First thing we do is try to load a file from the local filesystem.
 char LocalName[MAX_PATH];
 strcpy (LocalName , RootDir);
 strcat (LocalName , Name);

 HandleList[Handle].RealFile = CreateFile (LocalName , GENERIC_READ ,
                                           FILE_SHARE_READ , NULL ,
                                           OPEN_EXISTING ,
                                           FILE_FLAG_RANDOM_ACCESS , NULL);

 // If the file wasn't opened, then no file like that exists. If that is the
 // case, we have to try and find a file in a loaded NFI file that will match.
 if (HandleList[Handle].RealFile == INVALID_HANDLE_VALUE) {
     // Search for an NFI file with an entry that is like the one we want. If
     // there is one, set up the Handle information and return the handle.
     // Otherwise return -1.
     NFIDirEntry Entry;
     NFIFile     EntryFile;

     // Try to search for an entry with the right handle. If we don't find one
     // return -1.
     if (!SearchForNFIEntry (Entry , EntryFile , Name))
         return -1;

     // It worked! Entry and EntryFile are set as needed. So Set up the Handle
     // information and then return it.
     HandleList[Handle].RealFile = EntryFile.FileHandle;
     HandleList[Handle].IsPacked = TRUE;
     HandleList[Handle].StartPos = Entry.Offset;
     HandleList[Handle].Length = Entry.Length;
     HandleList[Handle].Position = 0;
     return Handle;
 }

 // If we get out here it means that there is a real file and it was opened, so
 // set up the rest of the information for the handle and then return it.
 HandleList[Handle].IsPacked = FALSE;
 HandleList[Handle].StartPos = 0;
 HandleList[Handle].Length = GetFileSize (HandleList[Handle].RealFile , NULL);
 HandleList[Handle].Position = 0;

 // Return the handle now.
 return Handle;
}

/*****************************************************************************
 * Name         - FSYSCloseFile
 * Purpose      - Closes a file which was opened using the FSYSOpenFile
 *                functions.
 * Arguments    - Handle: Datahandle associated with the file to close
 * Return Value - None
 ****************************************************************************/
void FSYSCloseFile (DATAHANDLE Handle)
{
 // Don't do anything if the handle isn't valid
 if (Handle < 0 || Handle >= MAX_DATA_HANDLES)
     return;

 if (HandleList[Handle].RealFile == INVALID_HANDLE_VALUE)
     return;

 // If this file is packed all we have to do to close it is set the RealFile
 // handle to NULL to mark this array entry as unused. There is no real file
 // to close until the program shuts down. Otherwise we close the handle
 // before marking it as closed.
 if (!HandleList[Handle].IsPacked)
     CloseHandle (HandleList[Handle].RealFile);

 HandleList[Handle].RealFile = INVALID_HANDLE_VALUE;
}

/*****************************************************************************
 * Name         - FSYSSeekFile
 * Purpose      - This functions seeks to a new location in the given file.
 *                The file is capable of seeking from either the start or the
 *                end of the file as the starting position. When seeking from
 *                the end use a negative position which indicates that you are
 *                seeking backwards. Currently there is no support for seeking
 *                from the current file position.
 * Arguments    - Handle: Local handle of the file to seek
 *                Position: New position to seek to
 *                Start: TRUE to seek from the start, FALSE to seek from the
 *                       end
 * Return Value - TRUE on success, FALSE on failure
 ****************************************************************************/
BOOL FSYSSeekFile (DATAHANDLE Handle , LONG Position , BOOL Start)
{
 // Don't do anything if the handle isn't valid
 if (Handle < 0 || Handle >= MAX_DATA_HANDLES)
     return FALSE;

 if (HandleList[Handle].RealFile == INVALID_HANDLE_VALUE)
     return FALSE;

 LONG Distance;

 // We want to convert all End file seeks to Begin file seeks. To do this we
 // Add up the file's start position and length, then add the position to that.
 // SInce the position should be negative, adding it actually subtracts it.
 // We now have an offset from the start of the file to seek to. This process
 // will work for both a regular file and a file system file since a regular
 // file has a 0 in it's startpos field which effectivly cancells out the
 // mechanism used to position for the start of a file in the filesystem.
 //
 // We also set the Postion member of the handle information to be the position
 // that was requested to be seeked to. The position is always 0 based and indexed
 // fro the start of the file, even if the file is packed. This is used so that
 // we can read from 2 files in an NFI File simultaneously if we want to without
 // messing things up. We need to keep a virtual record of where the file pointer
 // in any packed file might be.
 if (Start) {
     Distance = HandleList[Handle].StartPos + Position;
     HandleList[Handle].Position = Position;
 } else {
     Distance = HandleList[Handle].StartPos + HandleList[Handle].Length + Position;
     HandleList[Handle].Position = HandleList[Handle].Length + Position;
 }

 // Now do the seek operation.
 if (SetFilePointer (HandleList[Handle].RealFile , Distance , NULL ,
                     FILE_BEGIN) == 0xFFFFFFFF)
     return FALSE;

 // If we get here it worked.
 return TRUE;
}

/*****************************************************************************
 * Name         - FSYSReadByteStream
 * Purpose      - This reads a stream of byte data from the given handle into
 *                the provided buffer, which must be large enough to accept
 *                the data.
 * Arguments    - Handle: File to read from
 *                Buffer: Buffer to store the read data to
 *                Bytes: Number of bytes to read
 * Return Value - TRUE on success, FALSE on error
 ****************************************************************************/
BOOL FSYSReadByteStream (DATAHANDLE Handle , unsigned char *Buffer , DWORD Bytes)
{
 // Don't do anything if the handle isn't valid
 if (Handle < 0 || Handle >= MAX_DATA_HANDLES)
     return FALSE;

 if (HandleList[Handle].RealFile == INVALID_HANDLE_VALUE)
     return FALSE;

 DWORD BytesRead;

 // If the file is packed, there is a chance that a read from another file
 // in the packed file was done, which means the file pointer isn't in the
 // right position. So, if that is the case, do a seek before we read.
 if (HandleList[Handle].IsPacked)
     FSYSSeekFile (Handle , HandleList[Handle].Position , TRUE);

 // Read the data.
 BOOL Worked = (ReadFile (HandleList[Handle].RealFile , (LPVOID) Buffer ,
                         Bytes , &BytesRead , NULL) &&
                         (BytesRead == Bytes));

 HandleList[Handle].Position += BytesRead;

 return Worked;
}

/*****************************************************************************
 * Name         - FSYSReahShortByte
 * Purpose      - Reads a character from the given file
 * Arguments    - Handle: Handle to read from
 *                Character: Buffer to store the read data
 * Return Value - TRUE on success, FALSE on error.
 ****************************************************************************/
BOOL FSYSReadByte (DATAHANDLE Handle , unsigned char &Character)
{
 // Don't do anything if the handle isn't valid
 if (Handle < 0 || Handle >= MAX_DATA_HANDLES)
     return FALSE;

 if (HandleList[Handle].RealFile == INVALID_HANDLE_VALUE)
     return FALSE;

 DWORD BytesRead;

 // If the file is packed, there is a chance that a read from another file
 // in the packed file was done, which means the file pointer isn't in the
 // right position. So, if that is the case, do a seek before we read.
 if (HandleList[Handle].IsPacked)
     FSYSSeekFile (Handle , HandleList[Handle].Position , TRUE);

 // Read the data.
 BOOL Worked = (ReadFile (HandleList[Handle].RealFile , (LPVOID) &Character ,
                         sizeof (Character) , &BytesRead , NULL)) &&
               (BytesRead == sizeof (Character));


 HandleList[Handle].Position += BytesRead;

 return Worked;
}

/*****************************************************************************
 * Name         - FSYSReahShortInt
 * Purpose      - Reads a short integer from the given file
 * Arguments    - Handle: Handle to read from
 *                Integer: Buffer to store the read data
 * Return Value - TRUE on success, FALSE on error.
 ****************************************************************************/
BOOL FSYSReadShortInt (DATAHANDLE Handle , short int &Integer)
{
 // Don't do anything if the handle isn't valid
 if (Handle < 0 || Handle >= MAX_DATA_HANDLES)
     return FALSE;

 if (HandleList[Handle].RealFile == INVALID_HANDLE_VALUE)
     return FALSE;

 DWORD BytesRead;

 // If the file is packed, there is a chance that a read from another file
 // in the packed file was done, which means the file pointer isn't in the
 // right position. So, if that is the case, do a seek before we read.
 if (HandleList[Handle].IsPacked)
     FSYSSeekFile (Handle , HandleList[Handle].Position , TRUE);

 // Read the data.
 BOOL Worked = (ReadFile (HandleList[Handle].RealFile , (LPVOID) &Integer ,
                         sizeof (Integer) , &BytesRead , NULL)) &&
               (BytesRead == sizeof (Integer));

 HandleList[Handle].Position += BytesRead;

 return Worked;
}

/****************************************************************************/
